package com.sos.jobnet.context;

import com.sos.jobnet.db.JobNetPlanDBItem;
import com.sos.jobnet.db.JobNetPlanDBLayer;
import com.sos.jobnetmodel.objects.NodeStatus;
import com.sos.scheduler.cmd.JSKillTask;
import com.sos.scheduler.cmd.JSRemoveOrder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @uthor ss
 * at 09.09.13 16:41
 */
public class ContextList {

    private final static Logger logger = LoggerFactory.getLogger(ContextList.class);

    private final static Boolean SKIP_RUNNER = true;
    private final List<JobNetPlanDBItem> nodesInContext;
    private final JobNetPlanDBItem startNode;
    private final Map<NodeStatus,Integer> statusCounter = new HashMap<NodeStatus,Integer>();
    private final List<JobNetPlanDBItem> nodesToStart = new ArrayList<JobNetPlanDBItem>();
    private final JobNetPlanDBLayer planDBLayer;

    public ContextList(JobNetPlanDBLayer planDBLayer, List<JobNetPlanDBItem> nodesInContext, JobNetPlanDBItem startNode) {
        this.planDBLayer = planDBLayer;
        this.nodesInContext = nodesInContext;
        this.startNode = startNode;
        this.nodesToStart.add(startNode);           // default is to start the bootstrap node
        setStatusCounter();
    }

    private void setStatusCounter() {
        for(JobNetPlanDBItem item : nodesInContext) {
            NodeStatus status = NodeStatus.valueOf(item.getStatusNode());
            Integer count = (statusCounter.containsKey(status)) ? statusCounter.get(status) : 0;
            count++;
            statusCounter.put(status,count);
        }
    }

    public boolean isEmpty() {
        return nodesInContext.size() == 0;
    }

    public JobNetPlanDBItem getStartNode() {
        return startNode;
    }

    public List<JobNetPlanDBItem> getNodesInContext() {
        return nodesInContext;
    }

    public List<JobNetPlanDBItem> getNodesToStart() {
        return nodesToStart;
    }

    public Integer countNodesWithStatus(NodeStatus status) {
        return (statusCounter.containsKey(status)) ? statusCounter.get(status) : 0;
    }

    public boolean hasNodesWithStatus(NodeStatus status) {
        return (countNodesWithStatus(status) > 0);
    }

    private void changeStatus(NodeStatus fromStatus, NodeStatus toStatus) {
        int cnt = 0;
        for(JobNetPlanDBItem item : nodesInContext) {
            NodeStatus status = NodeStatus.valueOf(item.getStatusNode());
            if(status == fromStatus) {
                item.setStatusNode(toStatus.getIndex());
                planDBLayer.update(item);
            }
        }
        logger.info(cnt + " records in context changed from status " + fromStatus.name() + " to status " + toStatus.name() + ".");
    }

    /**
     * Submit command remove_order to all orders with status ERROR in the context. After that the affected orders will have the status KILLED.
     * @param host
     * @param port
     */
    public void resumeOrders(String host, Integer port) {
        nodesToStart.clear();
        int cnt = 0;
        for(JobNetPlanDBItem item : nodesInContext) {
            NodeStatus status = NodeStatus.valueOf(item.getStatusNode());
            if(status == NodeStatus.ERROR) {
                JSRemoveOrder remove = new JSRemoveOrder(host,port,item.getOrderId(), item.getJobChain());
                remove.run();
                item.setStatusNode(NodeStatus.KILLED.getIndex());
                planDBLayer.update(item);
                nodesToStart.add(item);
                cnt++;
                logger.info("Order " + item.getJobNetNodeDBItem().getNode() + " will be aborted.");
            }
        }
        logger.info(cnt + " records in context changed from status " + NodeStatus.ERROR.name() + " to status " + NodeStatus.KILLED.name() + ".");
    }

    /**
     * Submit command kill_task to all orders with status RUNNING in the context. After that the affected orders will have the status KILLED.
     * @param host
     * @param port
     */
    public void killTasks(String host, Integer port) {
        nodesToStart.clear();
        int cnt = 0;
        for(JobNetPlanDBItem item : nodesInContext) {
            NodeStatus status = NodeStatus.valueOf(item.getStatusNode());
            if(status == NodeStatus.RUNNING) {
                Integer taskId = item.getTaskId().intValue();
                String jobName = item.getJobName();
                JSKillTask kill = new JSKillTask(host,port,taskId, jobName, true);
                kill.run();
                item.setStatusNode(NodeStatus.KILLED.getIndex());
                planDBLayer.update(item);
                nodesToStart.add(item);
                cnt++;
                logger.info("Order " + item.getJobNetNodeDBItem().getNode() + " will be " + NodeStatus.KILLED.name() + ".");
            }
        }
        logger.info(cnt + " records in context changed from status " + NodeStatus.RUNNING.name() + " to status " + NodeStatus.KILLED.name() + ".");
    }

    /**
     * Submit command kill_task to all orders with status RUNNING in the context.
     * All order which are not FINISHED or IGNORED will be set to ABORTED.
     *
     * @param host
     * @param port
     */
    public void abortTasks(String host, Integer port) {
        nodesToStart.clear();
        int cnt = 0;
        for(JobNetPlanDBItem item : nodesInContext) {
            NodeStatus status = NodeStatus.valueOf(item.getStatusNode());
            if(status != NodeStatus.FINISHED && status != NodeStatus.IGNORED) {
                if(status == NodeStatus.RUNNING || status == NodeStatus.ERROR) {
                    if(status == NodeStatus.RUNNING) {
                        Integer taskId = item.getTaskId().intValue();
                        String jobName = item.getJobName();
                        JSKillTask kill = new JSKillTask(host,port,taskId, jobName, true);
                        kill.run();
                        logger.info("Task " + taskId + " (Job " + jobName + ") will be killed.");
                    }
                    JSRemoveOrder remove = new JSRemoveOrder(host,port,item.getOrderId(),item.getJobChain());
                    remove.run();
                    logger.info("Order " + item.getOrderId() + " will be removed.");
                }
                item.setStatusNode(NodeStatus.ABORTED.getIndex());
                planDBLayer.update(item);
                nodesToStart.add(item);
                cnt++;
                logger.info("Order " + item.getJobNetNodeDBItem().getNode() + " will be " + NodeStatus.ABORTED.name() + ".");
            }
        }
        logger.info(cnt + " records in context changed to status " + NodeStatus.ABORTED.name() + ".");
    }

    /**
     * All nodes marked with state ERROR or state KILLED will be prepared for restart. It means, the status will be changed to RESTART_PREPARED.
     * Nodes which already have the state RESTART_PREPARED will be placed in the result list, too.
     * Additionally the flag skipRunner is set. This causes to ignore the error next time.
     */
    public void ignoreErrors() {
        prepareErrors(SKIP_RUNNER);
    }

    /**
     * All nodes marked with state ERROR or state KILLED will be prepared for restart. That mean, the status will be changed to RESTART_PREPARED.
     * Nodes which already have the state RESTART_PREPARED will be placed in the result list, too.
     */
    public void prepareErrors() {
        prepareErrors(!SKIP_RUNNER);
    }

    private void prepareErrors(Boolean skipRunner) {
        nodesToStart.clear();
        int cnt = 0;
        for(JobNetPlanDBItem item : nodesInContext) {
            NodeStatus status = NodeStatus.valueOf(item.getStatusNode());
            if(status == NodeStatus.ERROR || status == NodeStatus.KILLED) {
                item.setStatusNode(NodeStatus.RESTART_PREPARED.getIndex());
                if (skipRunner) item.setIsRunnerSkipped(skipRunner);          // this causes to ignore the error if true
                planDBLayer.update(item);
                nodesToStart.add(item);
                cnt++;
                logger.info("Order " + item.getJobNetNodeDBItem().getNode() + " will be start again.");
            }
            if(status == NodeStatus.RESTART_PREPARED) {
                nodesToStart.add(item);
                logger.info("Order " + item.getJobNetNodeDBItem().getNode() + " is already marked to restart.");
            }
        }
        logger.info(cnt + " records in context changed from status " + NodeStatus.ERROR.name() + " to status " + NodeStatus.RESTART_PREPARED.name() + ".");
        logger.info(nodesToStart.size() + " records are prepared for restart now.");
        logger.info("The runners will be skipped next time.");
    }

    public void prepareContextForRestart() {
        int cnt = 0;
        for(JobNetPlanDBItem item : nodesInContext) {
            NodeStatus status = NodeStatus.valueOf(item.getStatusNode());
            if(status == NodeStatus.FINISHED || status == NodeStatus.ERROR) {
                item.setStatusNode(NodeStatus.RESTART_PREPARED.getIndex());
                item.setStatusWaiter(NodeStatus.NOT_PROCESSED.getIndex());
                item.setStatusRunner(NodeStatus.NOT_PROCESSED.getIndex());
                item.setStatusDispatcher(NodeStatus.NOT_PROCESSED.getIndex());
                planDBLayer.update(item);
                cnt++;
                logger.info("Order " + item.getJobNetNodeDBItem().getNode() + " will be start again.");
            }
            if(status == NodeStatus.RESTART_PREPARED) {
                logger.info("Order " + item.getJobNetNodeDBItem().getNode() + " is already marked to restart.");
            }
        }
        logger.info(cnt + " records in context changed to status " + NodeStatus.RESTART_PREPARED.name() + ".");
    }

}
