package com.sos.jobnet.creator;

import com.sos.JSHelper.Basics.JSToolBox;
import com.sos.jobnet.classes.JobNetConstants;
import com.sos.jobnet.classes.JobNetException;
import com.sos.jobnet.db.JobNetDBItem;
import com.sos.jobnet.db.JobNetEdgesDBItem;
import com.sos.jobnet.db.JobNetNodeDBItem;
import com.sos.jobnet.options.JobNetCreatorOptions;
import com.sos.jobnet.options.StartModes;
import com.sos.localization.Messages;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class is for creating a jobnet. A jobnet contains of a list of nodes connected by edges. Each jobnet has a special node called
 * bootstrap node (bootstrap order). The bootstrap node marks the starting point of the jobnet.
 * 
 * By calling the run method the complete jobnet (a collection of JobScheduler orders) will read from the filesystem and stored in the
 * database in the tables SCHEDULER_JOB_NET, SCHEDULER_JOB_NET_NODES and SCHEDULER_JOB_NET_EDGES.
 *
 * @author ss
 */
public class JobNetCreator implements Runnable, IDBConnection {
	
	@SuppressWarnings("unused")
	private static final String conSVNVersion = "$Id: JobNetCreator.java 22344 2014-02-03 13:22:47Z ss $";
	
	private final static Logger logger = LoggerFactory.getLogger(JobNetCreator.class);

	private static final Messages msg = new JSToolBox(JobNetConstants.strBundleBaseName).getMessageObject();
	private static final String	JOBNETC_I_0001 = "JOBNETC_I_0001";		// create job net for order %1$s (jobchain %2$s).
	private static final String JOBNETC_I_0002 = "JOBNETC_I_0002"; 		// %1$s nodes readed from file system.
	private static final String JOBNETC_I_0003 = "JOBNETC_I_0003";		// jobnet successfully written in database.
	private static final String JOBNETC_I_0005 = "JOBNETC_I_0005";		// the texture of the jobnet will not be checked.
	private static final String JOBNETC_I_0006 = "JOBNETC_I_0006";		// checking the texture of the jobnet.

	private final static String JOBNETC_E_0014 = "JOBNETC_E_0014"; 		// error creating jobnet
	
	private final JobNetCreatorOptions options;

	private final String orderId;
	private final String jobChain;
	private final String clusterId;
	private final boolean checkJobNetTexture;
    private final StartModes startOption;

    private JobNetBootstrapOrder bootstrapOrder;
    private long jobnetId;

	public JobNetCreator(JobNetCreatorOptions options) {
		this.options = options;
		this.orderId = options.getOrderIdWithPath();
		this.jobChain = options.getJobChainNameWithoutPath();
		this.clusterId = options.SCHEDULER_ID.Value();
		this.checkJobNetTexture = options.CheckJobNetTexture.value();
        this.startOption = options.StartOption.asEnum();
	}

	@Override
	public void run() {
		
        if(startOption.newInstance())
            convertConfigFilesToDB();
        else
            logger.info("For " + startOption.name() + " it is not necessary to create a new representation of the Jobnet in the datatabase.");

	}

    public StartModes getStartOption() {
        return startOption;
    }

	public void convertConfigFilesToDB() {

        logger.info(msg.getMsg(JOBNETC_I_0001,orderId,jobChain));
        bootstrapOrder = new JobNetBootstrapOrder(options);
        if(startOption == StartModes.start && !bootstrapOrder.isBootstrap())
            throw new JobNetException("The StartOption " + startOption.name() + " is valid for bootstrap orders only.");
        bootstrapOrder.createJobNet();

        if (checkJobNetTexture) {
            logger.info(msg.getMsg(JOBNETC_I_0006));
            bootstrapOrder.validateJobNet();
        } else {
            logger.info(msg.getMsg(JOBNETC_I_0005));
        }

        logger.info(msg.getMsg(JOBNETC_I_0002,getBootstrapOrder().getJobNet().size()));
        jobnetId = writeToDB();

        logger.info(msg.getMsg(JOBNETC_I_0003));

	}

	private long writeToDB() {

        long netId = 0;
        JobNetConifigurationLayer layer = options.getDbConfigurationLayer();
		try {
            layer.beginTransaction();

			// Create (or update) one record for each order in table SCHEDULER_JOB_NET_NODES
			for(JobNetOrder o : bootstrapOrder.getJobNet().values()) {
				JobNetOrderDBAdapter adapter = new JobNetOrderDBAdapter(layer, o);
                JobNetNodeDBItem nodeItem = adapter.createJobNetNode();
                o.setNodeDBItem(nodeItem);
                logger.debug("Create record with ID " + nodeItem.getNodeId());
			}
	
			// Create the edges for the jobnet in table SCHEDULER_JOB_NET_EDGES
			JobNetEdgesDBItem edgeItem;
			for(JobNetOrder parent : bootstrapOrder.getJobNet().values()) {
				JobNetOrderDBAdapter parentAdapter = new JobNetOrderDBAdapter(layer, parent);
				if (parentAdapter.itemExists()) {
					for(JobNetOrder child : parent.getSuccessors()) {
						JobNetOrderDBAdapter childAdapter = new JobNetOrderDBAdapter(layer, child);
						if (childAdapter.itemExists()) {
                            logger.debug("Create edges for " + parentAdapter.getItem().getNode() + " - " + childAdapter.getItem().getNode());
							parentAdapter.createSuccessor(childAdapter.getItem());
							childAdapter.createPredecessor(parentAdapter.getItem());
						}
					}
				}
			}

            // Create (or update) one record for the whole jobnet in table SCHEDULER_JOB_NET
            JobNetOrderDBAdapter jobnetAdapter = new JobNetOrderDBAdapter(layer, bootstrapOrder);
            JobNetDBItem jobnetItem = jobnetAdapter.createJobNet();
            netId = jobnetItem.getNetId();

			layer.commit();
		} finally {
		}
	    return netId;
    }

	public JobNetCreatorOptions getOptions() {
		return options;
	}

	public JobNetBootstrapOrder getBootstrapOrder() {
		return bootstrapOrder;
	}

	public long getNetId() {
		return jobnetId;
	}


    @Override
    public void closeSession() {
        options.getDbConfigurationLayer().closeSession();
    }
}
