package com.sos.jobnet.creator;

import com.sos.JSHelper.Basics.JSToolBox;
import com.sos.JSHelper.Exceptions.JobSchedulerException;
import com.sos.jobnet.classes.JobNetConstants;
import com.sos.jobnet.classes.JobNetNodeOrderStarter;
import com.sos.jobnet.db.JobNetPlanDBItem;
import com.sos.jobnet.interfaces.IJobSchedulerAdapter;
import com.sos.jobnet.interfaces.IMessageListener;
import com.sos.jobnet.logging.Level;
import com.sos.jobnet.options.JobNetCreatorOptions;
import com.sos.jobnet.options.StartBehaviour;
import com.sos.localization.Messages;
import com.sos.scheduler.cmd.ICommandActionsListener;
import com.sos.scheduler.model.objects.JSObjOrder;
import com.sos.scheduler.model.objects.JSObjParams;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * Creates a bootstrap order for every jobnet created with JobNetPlanCreator
 * 
 * The JobNetPlan creation for the given bootstrap order, which is called first in the run method, results
 * one or more jobnet plans. For each plan an order is created in the JobScheduler by sending the
 * command ADD_ORDER. 
 *
 * \author Stefan Schaedlich
 * \version 1.0 - 03.04.2012 10:53:07
 */
public class JobNetPlanExecutor implements IMessageListener, ICommandActionsListener {
	
	@SuppressWarnings("unused")
	private static final String conSVNVersion = "$Id$";
	
	private static final Messages msg = new JSToolBox(JobNetConstants.strBundleBaseName).getMessageObject();
//	private final static String JOBNETC_E_0010 = "JOBNETC_E_0010"; // order not started: %1$s.
//	private final static String JOBNETC_I_0011 = "JOBNETC_I_0011"; // order %1$s scheduled for %2$s.
	private final static String JOBNETC_E_0016 = "JOBNETC_E_0016"; // error creating the order for job net plan

	private final static Logger logger = LoggerFactory.getLogger(JobNetPlanExecutor.class);
//	private final static String YES = "yes";
	
//	private final ObjectModelConnector connection;
//	private final SchedulerObjectFactory factory;

//	private final String jobnetPrefix;
	private final String jobnetId;
	private final String jobnetIdQualifier;
//	private final String uuid;
	private final String host;
	private final int port;
	private final String jobChain;
	private final boolean apiStart;

	private List<String> commands = new ArrayList<String>();
	
	private IJobSchedulerAdapter orderStarter = null;
	
	public JobNetPlanExecutor(JobNetCreatorOptions options) {
		this.host = options.SCHEDULER_HOST.Value();
		this.port = options.SCHEDULER_TCP_PORT.value();
//		connection = ObjectModelConnector.getInstance(options.SCHEDULER_HOST, options.SCHEDULER_TCP_PORT, options.SCHEDULER_CONFIGURATION_DIRECTORY);
//		factory = connection.getFactory();
		this.jobnetIdQualifier = options.JobNetIdQualifier.Value();
		this.jobnetId = options.JobNetId.Value() + jobnetIdQualifier;
		//this.jobnetId = options.JobNetId.Value();
		this.jobChain = options.getJobChainNameWithoutPath();
		this.apiStart = options.APIStart.value();
//		this.jobnetPrefix = options.JobChainName.Value() + ".";
//		this.uuid = jobnetPrefix + jobnetId;
		String conMethodName = JobNetPlanExecutor.class.getSimpleName();
		logger.debug(msg.getMsg(JobNetConstants.JOBNET_D_0001, conMethodName, "JobNetId", jobnetId));
//		logger.debug(msg.getMsg(JobNetConstants.JOBNET_D_0001, conMethodName, "JobNetPrefix", jobnetPrefix));
//		logger.debug(msg.getMsg(JobNetConstants.JOBNET_D_0001, conMethodName, "UUID", uuid));
	}

	public void startOrder(JobNetPlanDBItem bootDBItem, StartBehaviour startBehaviour) {
		if (orderStarter != null)
			orderStarter.startOrder(bootDBItem, startBehaviour);
		else
			logger.warn("No object of type IJobSchedulerAdapter is registered to start an order.");
		//startOrder(bootDBItem, false);
	}
	
	public void startOrderViaJobScheduler(JobNetPlanDBItem planDBItem, StartBehaviour startBehaviour) {
		
		try {

            // String scriptRoot = orderStarter.getSpooler().variables().value(JobNetConstants.SCRIPT_ROOT_GLOBAL_PARAMETER_NAME);
			JobNetNodeOrderStarter starter = new JobNetNodeOrderStarter(host,port,planDBItem);

			JSObjOrder orderFromDB = new JSObjOrder(starter.getFactory());
			orderFromDB.getOrderFromXMLString(planDBItem.getOrderXml());
            JSObjParams paramsFromDB = new JSObjParams(starter.getFactory());
            paramsFromDB.setParamsFromString(planDBItem.getParamXml());

			// String id = orderFromDB.getId();
			String title = (orderFromDB.getTitle() == null) ? "<no title>" : orderFromDB.getTitle();

            starter.setState(orderFromDB.getState());
			starter.addMessageListener(this);
			starter.setReplace(true);
			starter.setTitle(title);
			starter.setAt((startBehaviour == StartBehaviour.at_runtime) ? getStartTime(planDBItem) : "now");
			starter.setParams(orderFromDB.getParams());
			starter.mergeParams(paramsFromDB);

            if (apiStart)
                starter.setSpooler(orderStarter.getSpooler());

            starter.run();
		} catch (Exception e) {
			String msgText = msg.getMsg(JOBNETC_E_0016);
			logger.error(msgText);
			throw new JobSchedulerException(msgText,e);
		}

	}
	
//	private void addParamToOrder(JSCmdAddOrder order, String name, String value) {
//		Param p = factory.createParam(name,value);
//		order.getParams().getParamOrCopyParamsOrInclude().add(p);
//		logger.info("Parameter set: " + name + "=" + value);
//	}
	
	private String getStartTime(JobNetPlanDBItem bootStrapDBItem) {
		return bootStrapDBItem.getPlannedStartTimeFormated();
	}

	public List<String> getCommands() {
		return commands;
	}
	
	@Override
	public void onMessage(Level level, String message) {
		Level.log(logger,level, message);
	}
	
	public void registerOrderStarter(IJobSchedulerAdapter orderStarter) {
		this.orderStarter = orderStarter;
	}

    @Override
    public void onMessage(org.apache.log4j.Level level, String message) {
        Level.log(logger, level, message);
    }
}
