package com.sos.jobnet.creator;

import com.sos.jobnet.classes.InternalJobNetError;
import com.sos.jobnet.classes.JobNetDAOConstants;
import com.sos.jobnet.classes.JobNetException;
import com.sos.jobnet.context.ContextBuilder;
import com.sos.jobnet.context.ContextList;
import com.sos.jobnet.db.*;
import com.sos.jobnet.interfaces.IJobSchedulerAdapter;
import com.sos.jobnet.options.JobNetCreatorOptions;
import com.sos.jobnet.options.SOSOptionStartJobnet;
import com.sos.jobnet.options.StartBehaviour;
import com.sos.jobnet.options.StartModes;
import com.sos.jobnetmodel.objects.NodeStatus;
import com.sos.scheduler.db.SchedulerInstancesDBItem;
import com.sos.scheduler.db.SchedulerInstancesDBLayer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;

/**
 * This is the entry point for processing a jobnet.
 */
public class JobNetProcessor extends JobNetPlanCreator implements Runnable, IJobSchedulerAdapter, IDBConnection {

	@SuppressWarnings("unused")
	private static final String conSVNVersion = "$Id: JobNetProcessor.java 21003 2013-09-06 08:13:52Z ss $";

	private final static Logger logger = LoggerFactory.getLogger(JobNetProcessor.class);

    private final SOSOptionStartJobnet startOption;
    private final StartModes startMode;

    private final String host;
    private final Integer port;
    private final String jobnetId;
    private final String jobnetIdQualifier;
    private final String jobChain;
    private final String orderIdWithPath;
    private final String orderId;
    private final String jobnetParameterName;
    private final String schedulerId;
    private final String externalGivenUUID;
    private final File configFile;
    private final JobNetPlanDBLayer planDbLayer;
    private final JobNetHistoryDBLayer historyDbLayer;
    private final JobNetCmdHistoryDBLayer cmdHistoryDbLayer;
    private final StartBehaviour startBehaviour;

    private JobNetHistoryDBItem historyRecord = null;
    private JobNetPlanDBItem startNode = null;
    private boolean uuidIsPresent = false;
    private ContextList contextList = null;
    private ContextBuilder contextBuilder = null;

    private boolean preconditionsChecked = false;
    private boolean databasePrepared = false;
    private String errorMessage = "";

	public JobNetProcessor(JobNetCreatorOptions options) {
		super(options);

        this.orderId = options.getNameComponents().getObjectName();
        this.jobChain = options.getJobChainNameWithoutPath();
        this.orderIdWithPath = options.getOrderIdWithPath();             // the order with leading live path

        this.startOption = options.StartOption;
        this.startMode = startOption.asEnum();
        this.jobnetIdQualifier = options.JobNetIdQualifier.Value();
        this.jobnetId = options.JobNetId.Value() + this.jobnetIdQualifier;
        //this.jobnetId = options.JobNetId.Value();
        this.jobnetParameterName = options.JobNetIdParameterName.Value();

        this.schedulerId = getOptions().SCHEDULER_ID.Value();
        this.configFile = new File(getOptions().hibernate_connection_config_file.Value());
        this.planDbLayer = options.getDbConfigurationLayer().getDbLayerPlan();
        this.historyDbLayer = options.getDbConfigurationLayer().getDbLayerHistory();
        this.cmdHistoryDbLayer = options.getDbConfigurationLayer().getDbLayerCommands();
        this.startBehaviour = options.StartBehaviour.asEnum();
        this.externalGivenUUID = jobChain + JobNetDAOConstants.JOBNET_ID_DELIMITER + jobnetId;

        this.host = options.SCHEDULER_HOST.Value();
        this.port = options.SCHEDULER_TCP_PORT.value();

        logger.info("Starting " + this.getClass().getSimpleName() + " with option " + startMode.name());
	}

    private void initContext(String uuid) {
        this.historyRecord = historyDbLayer.getJobNetInstance(uuid);
        if (historyRecord == null) {
            this.uuidIsPresent = false;
        } else {
            this.uuidIsPresent = true;
            contextBuilder = new ContextBuilder(uuid,orderId,planDbLayer);
            this.startNode = contextBuilder.getStartNode();
            this.contextList = contextBuilder.getContextList();
        }
    }

	@Override
	public void run() {

        String uuid = externalGivenUUID;
		try {

            logger.info("Try to initialize context.");
            initContext(uuid);

            // be sure that all preconditions are fulfilled
            if(!preconditionsChecked)
                checkPreconditions(uuid);

            // prepare the nodes depending on the StartOption
            if(!databasePrepared)
                prepareDatabase(uuid);

            // Set the context for the PlanCreator if available (the context is available at this thime only for those modes which works with an existing jobnet instance)
            if(!startMode.newInstance())
                if(contextList != null) setContext(contextList);

            // StartModes like kill or abort does not need to create a jobnet instance first.
            if (startMode.needCreator()) {

                super.run();

                // For those SartModes which creates a new jobnet instance the context is given from the PlanCreator
                if(startMode.newInstance()) {
                    logger.info("Get context from the new jobnet instance.");
                    this.historyRecord = getHistoryDBItem();
                    this.uuidIsPresent = true;
                    this.startNode = getStartNode();
                    this.contextList = getContext();
                }

                createJobNetCommandHistory();
                if (startBehaviour != StartBehaviour.no_start) {
                    logger.info("The start behaviour of the jobnet is " + startBehaviour.name());
                    for (JobNetPlanDBItem planItem : contextList.getNodesToStart()) {
                        logger.info(">>> Order to start: " + planItem.getOrderId());
                        startOrder(planItem, startBehaviour);
                    }
                } else {
                    setMessage("Jobnet not started because option StartBehaviour is set to " + startBehaviour.name() + ".");
                    logger.info(getMessage());
                }

            } else {

                // kill all running tasks in context
                createJobNetCommandHistory();
                if(startMode == StartModes.kill) {
                    contextList.killTasks(host,port);
                }

                if(startMode == StartModes.abort) {
                    contextList.abortTasks(host,port);
                    historyDbLayer.updateHistoryWithError(uuid, InternalJobNetError.JOBNET_ABORTED);
                }
            }

        } finally {
            logger.info("JobNetProcessor ended.");
		}

	}

    public void setMessage(String errorMessage) {
        this.errorMessage = errorMessage;
    }

    public String getMessage() {
        return errorMessage;
    }

    private void createJobNetCommandHistory() {
        if(historyRecord == null)
            logger.error("A history record for the jobnet with uuid " + externalGivenUUID + " is not set - could not create command history.");
        else {
            createJobNetCmdHistory();
        }
    }

	public void checkPreconditions(String uuid) {

        logger.info("Starting " + this.getClass().getSimpleName() + " with option " + startMode.name());

        if (startMode.checkUUID()) {
            if(!startMode.newInstance() && !isUUIDPresent() )
                throw new JobNetException("The StartOption " + startMode.name() + " needs an instance of the jobnet for JobNetId " + jobnetId + " - start a jobnet instance with StartOption 'start' first.");
            if(startMode.newInstance() && isUUIDPresent() )
                throw new JobNetException("An instance of the jobnet with JobNetId " + jobnetId + " is already started.");
        }

        if(startMode == StartModes.start) {
            logger.info("Need a new instance of a jobnet beginning with node " + orderIdWithPath + " - no preconditions to check.");
        }

        if(startMode == StartModes.start_independent) {
            logger.info("Need a new instance of a jobnet beginning with node " + orderIdWithPath + " - no preconditions to check.");
        }

        if(startMode == StartModes.repeat) {
            throwErrorIfNodesRunning();
            throwErrorIfJobnetEnded(uuid);
        }

        if(startMode == StartModes.restart_ignore_error) {
            throwErrorIfNodesRunning();
            throwErrorIfNoNodeInErrorOrRestartPrepared();
        }

        if(startMode == StartModes.restart_error) {
            throwErrorIfNodesRunning();
            throwErrorIfNoNodeInErrorOrRestartPrepared();
        }

        if(startMode == StartModes.kill) {
            throwErrorIfJobnetEnded(uuid);
        }

        if(startMode == StartModes.abort) {
            throwErrorIfJobnetEnded(uuid);
        }

        preconditionsChecked = true;
        logger.info("All preconditions to start the jobnet instance are fulfilled.");

	}

	public void prepareDatabase(String uuid) {

		try {

            planDbLayer.beginTransaction();

            if(!preconditionsChecked) {
                logger.info("Precondtions not checked - calling checkPreconditions first.");
                checkPreconditions(uuid);
            }

            if(startMode == StartModes.start) {
                logger.info("Need a new instance of a jobnet beginning with node " + orderIdWithPath + " - no database modifications are necessary.");
            }

            if(startMode == StartModes.start_independent) {
                logger.info("Need a new instance of a jobnet beginning with node " + orderIdWithPath + " - no database modifications are necessary.");
            }

            if(startMode == StartModes.repeat) {
                contextList.prepareContextForRestart();
                // contextList.resumeOrders(host,port);
            }

            if(startMode == StartModes.restart_ignore_error) {
                contextList.ignoreErrors();
            }

            if(startMode == StartModes.restart_error) {
                contextList.prepareErrors();
            }

            if(startMode == StartModes.kill) {
                logger.info("Killing tasks ...");
                contextList.killTasks(host, port);
            }

            if(startMode == StartModes.abort) {
                logger.info("StartOption " + startMode.name() + " not implemented yet.");
            }

            planDbLayer.commit();

            databasePrepared = true;
            logger.info("Database entries are prepared for the execution of the jobnet instance.");

		} finally {
        }

	}

    private JobNetCmdHistoryDBItem createJobNetCmdHistory() {
        return cmdHistoryDbLayer.createJobNetCmdHistory(startNode.getJobnetNodeDBItem().getNodeId(), historyRecord.getHistoryId(), startMode.name(), getMessage());
    }

    private void throwErrorIfNodesRunning() {
        if( contextList.hasNodesWithStatus(NodeStatus.RUNNING) ||
            contextList.hasNodesWithStatus(NodeStatus.WAITING) ||
            contextList.hasNodesWithStatus(NodeStatus.DISPATCHING) ||
            contextList.hasNodesWithStatus(NodeStatus.STARTED)
        )
            throw new JobNetException("Start with option " + startMode.name() + " not possible, because there are running nodes in the context of node " + startNode.getJobnetNodeDBItem().getNode());
    }

    private void throwErrorIfNoNodeInErrorOrRestartPrepared() {
        if(!contextList.hasNodesWithStatus(NodeStatus.ERROR) && !contextList.hasNodesWithStatus(NodeStatus.RESTART_PREPARED) && !contextList.hasNodesWithStatus(NodeStatus.KILLED))
            throw new JobNetException("StartOption " + startMode.name() + " ignored, because there are no nodes with status " + NodeStatus.ERROR.name() + " or " + NodeStatus.RESTART_PREPARED.name() + " or " + NodeStatus.KILLED.name() + " in the context.");
    }

    private void throwErrorIfJobnetEnded(String uuid) {
        if(historyRecord == null)
            throw new JobNetException("No history record for jobnet with uuid " + uuid + " is set.");
        if(historyDbLayer.isJobnetEnded(historyRecord))
            throw new JobNetException("Start with option " + startMode.name() + " not possible, because the jobnet instance with id " + uuid + " is already finished.");
    }

    private boolean isUUIDPresent() {
        return uuidIsPresent;
    }

    public void createSchedulerInstance() {

        if(getJSAdapter() == null)
            throw new JobNetException("Unable to create entry in scheduler_instances, because JS Adapter object not set (registerJSAdapter() not called).");

        SchedulerInstancesDBLayer instancesDBLayer = new SchedulerInstancesDBLayer(configFile);
        SchedulerInstancesDBItem record = new SchedulerInstancesDBItem();
        record.setSchedulerId(getJSAdapter().getSpooler().id());
        record.setHostName(getJSAdapter().getSpooler().hostname());
        record.setTcpPort(getJSAdapter().getSpooler().tcp_port());

        instancesDBLayer.beginTransaction();
        instancesDBLayer.saveOrUpdate(record);
        instancesDBLayer.commit();

    }

    /**
     * Read the startnode for a given jobnet instance
     * @return
    private JobNetPlanDBItem getStartNode() {
        JobNetPlanFilter startNodeFilter = new JobNetPlanFilter();
        logger.info("Search jobnet node with filter: schedulerID=" + schedulerId + ", uuid=" + historyRecord.getUuid() + ", oderID=" + orderIdWithPath);
        startNodeFilter.setSchedulerId(schedulerId);
        startNodeFilter.setUuid(historyRecord.getUuid());
        startNodeFilter.setNode(orderId);
        planDbLayer.setFilter(startNodeFilter);
        List<JobNetPlanDBItem> records = planDbLayer.getJobnetPlanList(0);
        if(records.size() != 1)
            throw new JobNetException("The orderIdWithPath " + orderIdWithPath + " is not unique in the jobnet instance with uuid " + uuid + ".");
        return records.get(0);
    }

    private ContextList getContextList(JobNetPlanDBItem startNode) {
        logger.info("Reading context nodes for order " + orderIdWithPath + " with UUID " + historyRecord.getUuid());
        List<JobNetPlanDBItem> contextNodes = planDbLayer.getMemberListForContext(startNode);
        ContextList list = new ContextList(planDbLayer, contextNodes,startNode);
        logger.info(list.getNodesInContext().size() + " nodes found in context.");
        for(NodeStatus status : NodeStatus.values())
            logger.info(list.countNodesWithStatus(status) + " nodes are in status " + status.name() + ".");
        return list;
    }
     */

}
