package com.sos.jobnet.db;

import com.sos.JSHelper.Exceptions.JobSchedulerException;
import com.sos.hibernate.layer.SOSHibernateDBLayer;
import com.sos.jobnet.classes.ConditionType;
import com.sos.jobnet.classes.JobNetDAOConstants;
import com.sos.jobnet.classes.JobNetDAOException;
import com.sos.jobnet.classes.ConditionState;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.List;

public class ConditionsDBLayer extends SOSHibernateDBLayer {

	private final static Logger logger = LoggerFactory.getLogger(ConditionsDBLayer.class);

	private final static String PLAN_ID = "planId";
	private final static String CONDITION_NAME = "conditionName";
	private final static String CONDITION_STATE = "conditionState";
	private final static String CONDITION_TYPE = "conditionType";

    private ConditionsFilter filter = new ConditionsFilter();

    public ConditionsDBLayer(File configurationFile) {
        super();
        initSession();
        this.setConfigurationFile(configurationFile);
	}

    public ConditionsFilter getFilter() {
        return filter;
    }

    public void setFilter(ConditionsFilter filter) {
        this.filter = filter;
    }

    private Query setQueryParams(String hql) {
		Query query = null;
		try {
			query = session.createQuery(hql);

            if (filter.getPlanId() != null)
		        query.setParameter(PLAN_ID, filter.getPlanId());

            if (filter.getConditionName() != null)
		        query.setParameter(CONDITION_NAME, filter.getConditionName());

            if (filter.getConditionType() != null)
		        query.setParameter(CONDITION_TYPE, filter.getConditionType());

            if (filter.getConditionState() != null)
		        query.setParameter(CONDITION_STATE, filter.getConditionState());

        } catch(HibernateException e) {
			throw new JobSchedulerException("Error creating Query",e);
		}
		return query;
	}

	private String getWhere() {
		String where = "";
		String and = "";

        if (filter.getPlanId() != null) {
            where += and + " planId = :planId";
            and = " and ";
        }

        if (filter.getConditionName() != null) {
            where += and + " conditionName = :conditionName";
            and = " and ";
        }

        if (filter.getConditionType() != null) {
            where += and + " conditionType = :conditionType";
            and = " and ";
        }

        if (filter.getConditionState() != null) {
            where += and + " conditionState = :conditionState";
            and = " and ";
        }

		return (where.isEmpty()) ? where : "where " + where;

	}

    public List<ConditionsDBItem> getJobNetConditionsList(int limit) {

        initSession();
        Query query = setQueryParams("from ConditionsDBItem conditions " + getWhere() + this.filter.getOrderCriteria() + this.filter.getSortMode());
        if (limit > 0) {
            query.setMaxResults(limit);
        }

        List<ConditionsDBItem> conditionsList = query.list();
        return conditionsList;
    }

    public List<ConditionsDBItem> getConditions(Long planId) {
        filter = new ConditionsFilter();
        filter.setPlanId(planId);
        return getJobNetConditionsList(0);
    }

    public List<ConditionsDBItem> getConditions(Long planId, ConditionState state) {
        filter = new ConditionsFilter();
        filter.setPlanId(planId);
        filter.setConditionState(state);
        return getJobNetConditionsList(0);
    }

    public List<ConditionsDBItem> getMissingConditions(Long planId) {
        return getConditions(planId, ConditionState.MISSING);
    }

    /**
     * Delete all entries in the table.
     * @return
     */
    public int delete() {
        initSession();
        Query query = session.createQuery("delete from ConditionsDBItem");
        return query.executeUpdate();
    }

    /**
     * A condition is uniquely identified by a node id and the name of the condition. The state of the specified
     * condition is changed to FULFILLED if it is not already in state IGNORED.
     *
     * @param planId
     * @param name
     */
    public void setConditionFulfilled(Long planId, String name) {
        filter = new ConditionsFilter();
        filter.setConditionState(ConditionState.MISSING);
        filter.setConditionName(name);
        filter.setPlanId(planId);

        List<ConditionsDBItem> conditions = getJobNetConditionsList(0);
        if(conditions.size() > 1)
            throw new JobNetDAOException("Condition " + name + " is not unique for planId " +planId);

        if (conditions.size() == 1) {
            ConditionsDBItem condition = conditions.get(0);
            updateRecord(condition, ConditionState.FULFILLED);
        }
    }

    /**
     * Sets all conditions with a given name to state FULFILLED if they have state CREATED or MISSING.
     *
     * @param name
     */
    public void setConditionFulfilled(String name) {
        filter = new ConditionsFilter();
        filter.setConditionName(name);

        List<ConditionsDBItem> conditions = getJobNetConditionsList(0);
        logger.info("Found " + conditions.size() + " missing conditions with name " + name);
        for(ConditionsDBItem condition : conditions) {
            if (condition.getConditionState() == ConditionState.CREATED || condition.getConditionState() == ConditionState.MISSING) {
                logger.debug("Update condition for node with plan_id " + condition.getPlanId());
                updateRecord(condition,ConditionState.FULFILLED);
            }
        }
    }

    public void createPredecessorCondition(JobNetPlanDBItem forNode, String predecessor) {
        ConditionsDBItem newRecord = createCondition(forNode.getPlanId(),predecessor,ConditionType.PREDECESSOR);
        updateRecord(newRecord, ConditionState.MISSING);
    }

    public void createSingleEventCondition(JobNetPlanDBItem forNode, String eventName) {
        ConditionsDBItem newRecord = createCondition(forNode.getPlanId(),eventName,ConditionType.EVENT);
        updateRecord(newRecord, ConditionState.CREATED);
    }

    public void createActiveSingleEventCondition(JobNetPlanDBItem forNode, String eventName) {
        ConditionsDBItem newRecord = createCondition(forNode.getPlanId(),eventName,ConditionType.EVENT);
        updateRecord(newRecord, ConditionState.MISSING);
    }

    public void createOnDemandCondition(JobNetPlanDBItem forNode) {
        ConditionsDBItem newRecord = createCondition(forNode.getPlanId(),ConditionType.ON_DEMAND.name(),ConditionType.ON_DEMAND);
        updateRecord(newRecord, ConditionState.MISSING);
    }

    /**
     * Switches all conditions for a node which are in state CREATED to state MISSING.
     *
     * @param forNode
     */
    public List<ConditionsDBItem> switchToMissingState(JobNetPlanDBItem forNode) {
        return switchState(forNode.getPlanId(), forNode.getNodeName(), ConditionState.CREATED, ConditionState.MISSING);
    }

    /**
     * Switches all events in a given list to state fulfilled if necessary
     *
     * @param forNode
     */
    public void setEventsFulfilled(JobNetPlanDBItem forNode, List<EventsDBItem> forEvents) {

        filter = new ConditionsFilter();
        filter.setPlanId(forNode.getPlanId());
        filter.setConditionState(ConditionState.MISSING);
        filter.setConditionType(ConditionType.EVENT);
        List<ConditionsDBItem> conditions = getJobNetConditionsList(0);
        logger.info(conditions.size() + " event conditions missing for node " + forNode.getNodeName());

        for(ConditionsDBItem condition : conditions) {
            logger.debug("event condition: " + condition.getConditionName());
            for(EventsDBItem event : forEvents) {
                logger.debug("compare with event: " + event.getEventId() + " resp. " + event.getEventAsString());
                if(event.isEqual(condition.getConditionName()))
                    updateRecord(condition,ConditionState.FULFILLED);
            }
        }
    }

    /**
     * Switches all conditions for a node which are in state MISSING to state SKIPPED.
     *
     * @param forNode
     */
    public List<ConditionsDBItem> switchToSkippedState(JobNetPlanDBItem forNode) {
        return switchState(forNode.getPlanId(), forNode.getNodeName(), ConditionState.MISSING, ConditionState.SKIPPED);
    }

    public static StringBuffer asStringBuffer(List<ConditionsDBItem> missingConditions) {
        StringBuffer msg = new StringBuffer();
        for(ConditionsDBItem condition : missingConditions) {
            if(msg.length() > 0) msg.append("; ");
            msg.append("[");
            msg.append(condition.getConditionType().getMsg());
            msg.append("] ");
            msg.append(condition.getConditionName());
            msg.append("\n");
        }
        return msg;
    }

    private ConditionsDBItem createCondition(Long planId, String condition, ConditionType type) {
        ConditionsDBItem newRecord = new ConditionsDBItem();
        newRecord.setPlanId(planId);
        newRecord.setConditionType(type);
        newRecord.setConditionName(condition);
        newRecord.setNote("waiting for " + type.getMsg() + " " + condition);
        newRecord.setMailToOnApprove("");
        return newRecord;
    }

    private List<ConditionsDBItem> switchState(Long planId, String nodeName, ConditionState fromState, ConditionState toState) {
        filter = new ConditionsFilter();
        filter.setPlanId(planId);
        filter.setConditionState(fromState);
        List<ConditionsDBItem> conditions = getJobNetConditionsList(0);

        if(conditions.size() > 0) {
            for(ConditionsDBItem condition : conditions) {
                if(toState.isMissing())
                    condition.setMissedFrom(new DateTime());
                if(toState.isFinished())
                    condition.setFinishedAt(new DateTime());
                updateRecord(condition,toState);
            }
            logger.info(conditions.size() + " conditions for node " + nodeName + " are changed from " + fromState.name() + " to " + toState.name() + " state.");
        } else {
            logger.info("Node Conditions for node " + nodeName + " are in state " + fromState.name() + " - nothing to do.");
        }
        return conditions;
    }

    private void updateRecord(ConditionsDBItem condition, ConditionState toState) {
        condition.setConditionState(toState);
        condition.setChangedBy(JobNetDAOConstants.CURRENT_USER);
        condition.setChangedAt(new DateTime());
        saveOrUpdate(condition);
    }
}
