package com.sos.jobnet.db;

import com.sos.JSHelper.Exceptions.JobSchedulerException;
import com.sos.hibernate.layer.SOSHibernateDBLayer;
import com.sos.jobnet.classes.EventsCollection;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class EventsDBLayer extends SOSHibernateDBLayer {

	private final static Logger logger = LoggerFactory.getLogger(EventsDBLayer.class);

	private final static String EVENT_ID = "eventId";
	private final static String EVENT_CLASS = "eventClass";
	
	public EventsDBLayer(File configurationFile) {
        super();
        this.setConfigurationFile(configurationFile);
	}
	
	public EventsDBLayer(File configurationFile, EventsFilter filter) {
        super();
        this.setConfigurationFile(configurationFile);
	}
	
    private Query setQueryParams(EventsFilter filter, String hql) {
		Query query = null;
		try {
			query = session.createQuery(hql);
			if (filter.hasEvents())
		        query.setParameterList(EVENT_ID, filter.getEventList());
			if (filter.hasEventClass())
		        query.setParameter(EVENT_CLASS, filter.getEventClass());
		} catch(HibernateException e) {
			throw new JobSchedulerException("Error creating Query",e);
		}
		return query;
	}

	private String getWhere(EventsFilter filter) {
		String where = "";
		String and = "";

        if (filter.hasEvents()) {
            where += and + " eventId in ( :eventId )";
            and = " and ";
        }

        if (filter.hasEventClass()) {
            where += and + " eventClass = :eventClass";
            and = " and ";
        }

		return (where.isEmpty()) ? where : "where " + where;

	}

	/**
	 * Checks the events in the eventList if present and returns a list of all events still missing.
     *
	 * @param eventList
	 * @return
	 */
	public EventsCollection checkEvents(List<String> eventList) {
        EventsCollection missingEvents = new EventsCollection();
        List<EventsDBItem> resultList = getFulfilledEvents(eventList);
        missingEvents.addAll(eventList);
        for(EventsDBItem item : resultList) {
            missingEvents.remove(item.getEventId());
            missingEvents.remove(getEventAsString(item));           // a fully addressed event: event_class/event_id
        }
    	logger.info("missing events " + missingEvents.toString());
		return missingEvents;
	}

	/**
	 * Checks the events in the eventList if present and returns a list of all fulfilled events.
     *
	 * @param eventList
	 * @return
	 */
	public List<EventsDBItem> getFulfilledEvents(List<String> eventList) {
        List<EventsDBItem> resultList = new ArrayList<EventsDBItem>();
        logger.info("Check the presence of the following events: " + eventList.toString());
        EventsCollection missingEvents = new EventsCollection();
		EventsFilter filter = new EventsFilter();
		filter.setEventList(eventList);
		if(filter.getEventList().isEmpty()) {
			logger.info("None of the given events is fulfilled.");
		} else {
            if (!getSession().isOpen()) initSession();
            logger.info("check events " + filter.getEventList().toString());
            Query query = setQueryParams(filter, "from EventsDBItem table_nodes "
                    + getWhere(filter));
            resultList = query.list();
        }
		return resultList;
	}

    public static String getEventAsString(EventsDBItem event) {
        return event.getEventClass() + "/" + event.getEventId();
    }
	
	public boolean isEventThrown(String eventId) {
		List<String> events = Arrays.asList(eventId);
		return (checkEvents(events).size() == 0);
	}
	
	public void createEvent(EventsDBItem record) {
        if(!isEventThrown(record.getEventId())) {
            DateTime now = new DateTime();
            DateTime expired = now.plusDays(60);
            record.setCreated( new DateTime() );
            record.setExpires(expired);
            saveOrUpdate(record);
        }
	}
	
	public void deleteEventsForClass(String eventClass) {
		EventsFilter filter = new EventsFilter();
		filter.setEventClass(eventClass);
    	if (!getSession().isOpen()) initSession();
		Query query = setQueryParams(filter, "from EventsDBItem table_nodes "
				+ getWhere(filter));
				// + this.filter.getOrderCriteria()
				// + this.filter.getSortMode());

		List<EventsDBItem> resultList = query.list();
		logger.info("Delete " + resultList.size() + " events for eventClass " + eventClass);
		for(EventsDBItem item : resultList) {
			delete(item);
		}
	}

	public void delete() {
    	if (!getSession().isOpen()) initSession();
		Query query = session.createQuery("from EventsDBItem table_nodes ");
		List<EventsDBItem> resultList = query.list();
		logger.info("Delete " + resultList.size() + " events.");
		for(EventsDBItem item : resultList) {
			delete(item);
		}
	}

}
