package com.sos.jobnet.jobs.dispatcher;

import com.sos.jobnet.classes.JobNetCollection;
import com.sos.jobnet.classes.JobNetException;
import com.sos.jobnet.db.EventsDBItem;
import com.sos.jobnet.db.JobNetCmdHistoryDBItem;
import com.sos.jobnet.db.JobNetPlanDBItem;
import com.sos.jobnet.interfaces.IDispatchHandler;
import com.sos.jobnet.interfaces.IJobNetCallback;
import com.sos.jobnet.interfaces.IJobNetPlanEventListener;
import com.sos.jobnet.jobs.waiter.JobNetWaiter;
import com.sos.jobnet.logging.Level;
import com.sos.jobnet.options.JobNetOptions;
import com.sos.jobnet.options.StartModes;
import com.sos.jobnetmodel.objects.NodeStatus;
import com.sos.scheduler.model.SchedulerObjectFactory;
import com.sos.scheduler.model.objects.JSObjParams;
import com.sos.scheduler.model.objects.Params;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.List;

/**
* Class for dispatching the successor nodes of a jobnet node.
* 
* @author Uwe Risse
* @version $Id: JobNetDispatcher.java 21059 2013-09-12 09:51:06Z ss $
*
* Created on 24.02.2012 13:02:19
 */
public class JobNetDispatcher implements IJobNetCallback, IJobNetPlanEventListener {

	public  final String	conSVNVersion	= "$Id: JobNetDispatcher.java 21059 2013-09-12 09:51:06Z ss $";
	private final String		conClassName		= "JobNetDispatcher";
	private static final Logger	logger				= LoggerFactory.getLogger(JobNetDispatcher.class);
	private IDispatchHandler	objDispatchHandler	= null;
	private JobNetOptions		objOptions			= null;

	private IJobNetCallback objJobNetCallback = this;

	public JobNetDispatcher() {
	}
	
	public void setIJobNetCallback(IJobNetCallback pobjJobNetCallback) {
		objJobNetCallback = pobjJobNetCallback;
	}
	
	public JobNetOptions Options() {
		if (objOptions == null)
			objOptions = new JobNetOptions();
		return objOptions;
	}

	public void setOptions(JobNetOptions objOptions) {
		this.objOptions = objOptions;
	}
	
	public void setOptionsUsingHashMap(HashMap<String, String> objHshMap) {
		try {
			Options().setAllOptions(objHshMap);
		}
		catch (Exception e) {
 			e.printStackTrace();
		}
	}

	public void doInit() {
		objDispatchHandler = new JobNetDBDispatcher(objOptions);
		objDispatchHandler.setNodeStatusListener(this);
	}

	public JobNetDispatcher run() {

		objDispatchHandler.open();
		boolean doDispatching = true;
        boolean jobnetEnded = false;

 		try {

            if (objDispatchHandler.isRunnerSkipped()) {
                objDispatchHandler.setRunnerStatusToFinished();
            } else {
                int exitCode = objJobNetCallback.getExitCode();
                objJobNetCallback.log(Level.INFO,"The exit_code of the script is " + exitCode);

                if (!Options().valid_errors.IsEmpty())
                    objJobNetCallback.log(Level.INFO,"Valid errors are " + Options().valid_errors.Value());

                if ( exitCode!=0 && !Options().valid_errors.contains(exitCode) ) {
                    logger.info("Error ignored, because it is in the list of valid errors.");
                    objDispatchHandler.setNodeStatusToError(exitCode);
                    doDispatching = false;
                    if ( Options().valid_errors.contains(exitCode) )
                        objDispatchHandler.setRunnerStatusToFinished();
                    else
                        objDispatchHandler.setRunnerStatusToError();
                }

            }

            if (doDispatching) {
                objDispatchHandler.setDispatcherStatusToRunning();
                objDispatchHandler.setNodeStatusToDispatching();
                dispatch();
                objDispatchHandler.setNodeStatusToFinished();
                objDispatchHandler.setDispatcherStatusToFinished();
                jobnetEnded = objDispatchHandler.updateHistoryNodeEnds();
                objDispatchHandler.setPredecessorConditionFulfilled();
            }

		} catch (Exception e) {
			objDispatchHandler.setDispatcherStatusToError();
			objDispatchHandler.setNodeStatusToError(0);
			String msg = "Error processing dispatcher.";
			logger.error(msg,e);
			throw new JobNetException(msg, e);

		} finally {
            if (jobnetEnded) {
                if (objDispatchHandler.workWithEvents()) objJobNetCallback.createJobNetEndedEvent(this);
                objJobNetCallback.setStateText("jobnet finished.");
            }
            objDispatchHandler.close();
            objJobNetCallback.log(Level.INFO,"Dispatcher is finished.");
		}
		return this;
	}

	public void createEvent(String eventId) {
		String uuid = Options().uuid_jobnet_identifier.Value();
        JobNetCmdHistoryDBItem command = objDispatchHandler.getCmdHistory();
        StartModes startMode = StartModes.valueOf(command.getStartOption());
        if(objDispatchHandler.workWithEvents()) {
            EventsDBItem eventRecord = objJobNetCallback.createEventRecord(uuid, eventId);
            objDispatchHandler.createEvent(eventRecord);
        } else {
            logger.info("Event " + eventId + " not thrown, because an independent instance is running.");
        }
	}

    /**
     * Dispatching all subsequent nodes (if not already dispatched by another node)
     */
    private void dispatch(){
 		
        if(!objDispatchHandler.isDispatcherSkipped()) {

            JobNetCollection jobNetCollection = objDispatchHandler.getDispatchingNodes();
            if (jobNetCollection.size() == 0) {
                logger.info("No nodes found to dispatch for order " + objDispatchHandler.getCurrentRecord().getJobnetNodeDBItem().getNode());
            } else {
                JSObjParams params = objJobNetCallback.getOrderParameterToProvide();
                JobNetPlanDBItem currentNode = objDispatchHandler.getCurrentRecord();
                if(currentNode.isSubnet()) {
                    supplyParameterToSubsequentConnectors(params);
                } else {
                    if(currentNode.isConnector()) {
                        jobNetCollection = objDispatchHandler.getDispatchingNodes();
                        supplyParameterToSubsequentNodes(jobNetCollection, params);
                    }
                }

                // read the collection again to get updated parameter values
                jobNetCollection = objDispatchHandler.getDispatchingNodes();
                for(int i = 0; i < jobNetCollection.size();i++){
                    JobNetPlanDBItem planItem = (JobNetPlanDBItem)jobNetCollection.get(i);
                    objJobNetCallback.startNode(planItem, Options());
                }
            }
        }
	}

    /**
     * Only the first connector node is defined as successors of a subnet, but all connectors must get the parameters of its parent subnet.
     */
    private void supplyParameterToSubsequentConnectors(JSObjParams params) {
            List<JobNetPlanDBItem> children = objDispatchHandler.getChildNodes();
            for(JobNetPlanDBItem child : children) {
                JSObjParams currentParams = objJobNetCallback.getFactory().createParams();
                if(child.getParamXml() != null)
                    currentParams = new JSObjParams (objJobNetCallback.getFactory(), (Params)currentParams.unMarshal(child.getParamXml()));
                setNodeParameter(child, currentParams, params);
            }
    }

    private void supplyParameterToSubsequentNodes(JobNetCollection subsequentNodes, JSObjParams paramsOfCurrentOrder) {
        for(int i = 0; i < subsequentNodes.size();i++){
            JobNetPlanDBItem planItem = (JobNetPlanDBItem)subsequentNodes.get(i);
            JSObjParams successorParams = objJobNetCallback.getFactory().createParams();
            if(planItem.getParamXml() != null)
                successorParams = new JSObjParams (objJobNetCallback.getFactory(), (Params)successorParams.unMarshal(planItem.getParamXml()));
            setNodeParameter(planItem,successorParams, paramsOfCurrentOrder);
        }
    }

    private void setNodeParameter(JobNetPlanDBItem item, JSObjParams successorParams, JSObjParams paramsOfCurrentOrder) {
        try {
            logger.info("Set parameter for " + item.getNodeType() + " " + item.getJobNetNodeDBItem().getNode());
            logger.trace("successorParams: " + successorParams.toXMLString());
            logger.trace("paramsOfCurrentOrder: " + paramsOfCurrentOrder.toXMLString());
            successorParams.merge(paramsOfCurrentOrder);
            objDispatchHandler.updateParams(item, successorParams);
            logger.trace("mergedParams: " + successorParams.toXMLString());
        } catch (Exception e) {
            String msg = "Error while setting parameters for node " + item.getJobnetNodeDBItem().getNode();
            logger.error(msg);
            throw new JobNetException(msg,e);
        }
    }

	@Override
	public int getExitCode() {
		// TODO Auto-generated method stub
		return 0;
	}

    @Override
    public void setNextState() {
        //To change body of implemented methods use File | Settings | File Templates.
    }

    @Override
	public void log(Level level, String message, Throwable error) {
		Level.log(logger,level, message, error);
	}

	@Override
	public void log(Level level, String message) {
		Level.log(logger,level, message);
	}

	@Override
	public void setStateText(String stateText) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void startNode(JobNetPlanDBItem planItem, JobNetOptions options) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void skipRunner() {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void logStatus(JobNetPlanDBItem record, NodeStatus oldStatus) {
        NodeStatus status = NodeStatus.valueOf(record.getStatusNode());
		logger.info("Node status of id " + record.getNodeId() + " is set from " + oldStatus.name() + " to " + status.name());
	}

	@Override
	public void nodeStatusChanged(JobNetPlanDBItem node, NodeStatus oldState) {
		objJobNetCallback.logStatus(node, oldState);
	}

	@Override
	public void waiterStatusChanged(NodeStatus newState, NodeStatus oldState) {
		// not implemented
	}

	@Override
	public void runnerStatusChanged(NodeStatus newState, NodeStatus oldState) {
		objJobNetCallback.logContextStatus("STATUS_RUNNER", oldState, newState);
	}

	@Override
	public void dispatcherStatusChanged(NodeStatus newState, NodeStatus oldState) {
		objJobNetCallback.logContextStatus("STATUS_DISPATCHER", oldState, newState);
	}

	@Override
	public void logContextStatus(String context, NodeStatus fromStatus, NodeStatus toStatus) {
		String msg = context + " changed from  " + fromStatus.name() +  " to " + toStatus.name();
		logger.info(msg);
	}

	@Override
	public EventsDBItem createEventRecord(String forEventClass,	String forEventId) {
		// has to be implemented in the callback class.
		return null;
	}

	@Override
	public void createJobNetStartedEvent(JobNetWaiter waiter) {
		// has to be implemented in the callback class.
	}

    @Override
    public JSObjParams getOrderParameterToProvide() {
        // has to be implemented in the callback class.
        return null;  //To change body of implemented methods use File | Settings | File Templates.
    }

    @Override
    public SchedulerObjectFactory getFactory() {
        return null;  //To change body of implemented methods use File | Settings | File Templates.
    }

    @Override
	public void createJobNetEndedEvent(JobNetDispatcher dispatcher) {
        // has to be implemented in the callback class.
    }
}
