package com.sos.jobnetmodel.live;

import com.sos.jobnetmodel.objects.ExternalJobnetNodeObject;
import com.sos.jobnetmodel.objects.JobnetNode;
import com.sos.jobnetmodel.objects.JobnetNodeImpl;
import com.sos.jobnetmodel.objects.JobnetNodeType;
import com.sos.scheduler.model.objects.JSObjOrder;
import com.sos.scheduler.model.objects.JSObjParams;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * This class is to convert the xml configuration of an order to an internal object. It is necessary that it implements the
 * minimal interface of a jobnet node (ExternalJobnetNodeObject).
 *
 * @uthor ss
 * at 13.03.13 16:16
 */
public class JSOrderAdapter implements ExternalJobnetNodeObject {

    private final Logger logger = LoggerFactory.getLogger(JSOrderAdapter.class);
    private final JSObjOrder order;
    private final String xmlContent;
    private final List<String> successors;
    private final List<String> predecessors;
    private final List<Integer> validErrors;
    private final boolean isBootstrapNode;
    private final JobnetNodeType type;
    private final String scriptToExecute;
    private final String restartOption;
    private final String eventOnError;
    private final String eventOnSuccess;
    private final boolean errorHandler;
    private final String subnet;
    private final String connector;
    private final String id;
    private final JobnetNode netNode;

    public JSOrderAdapter(JSObjOrder order) {
        this.order = order;
        this.xmlContent = order.toXMLString();
        this.successors = buildStringList("successor");
        this.predecessors = buildStringList("predecessor");
        this.isBootstrapNode = buildBoolean("bootstrap");
        this.errorHandler = buildBoolean("error_handler");
        this.validErrors = buildIntegerList("valid_errors");
        this.type = JobnetNodeType.valueOfWithException(buildString("nodetype"));
        this.scriptToExecute = buildString("script_to_execute");
        this.restartOption = buildString("restart_option");
        this.eventOnError = buildString("event_on_error");
        this.eventOnSuccess = buildString("event_on_success");
        this.subnet = buildString("subnet");
        this.connector = buildString("connector");
        this.id = getId();
        this.netNode = new JobnetNodeImpl(id, type);
        // setNodeProperties();
    }

    public List<Integer> getValidErrors() {
        return validErrors;
    }

    @Override
    public List<LockUsed> getUsedLocks() {
        return null;
    }

    private String getId() {
        String result = order.getId();
        if (result == null) {
            result = order.getHotFolderSrc().getName().replace(".order.xml","");
            int i = result.lastIndexOf(',');
            result = result.substring(i+1);
        }
        return result;
    }

    private List<Integer> buildIntegerList(String parameterName) {
        JSObjParams params = order.getJSObjParams();
        List<Integer> result = new ArrayList<Integer>();
        if (params.hasParameterValue(parameterName)) {
            String paramValue = params.values().get(parameterName);
            result.addAll(ExitCodeList.getExitCodes(paramValue));
        } else
            logger.info("Parameter " + parameterName + " not set in order " + order.getId() + ".");
        return result;
    }

    private List<String> buildStringList(String parameterName) {
        JSObjParams params = order.getJSObjParams();
        List<String> result = new ArrayList<String>();
        if (params.hasParameterValue(parameterName)) {
            String successors = params.values().get(parameterName);
            String[] arr = successors.split(",");
            for (String id : arr) {
                result.add(id);
            }
        } else
            logger.info("Parameter " + parameterName + " not set in order " + order.getId() + ".");
        return result;
    }

    private boolean buildBoolean(String parameterName) {
        JSObjParams params = order.getJSObjParams();
        boolean result = false;
        if (params.hasParameterValue(parameterName)) {
            String booleanAsString = params.values().get(parameterName);
            result = order.getYesOrNo(booleanAsString);
        } else
            logger.info("Parameter " + parameterName + " not set in order " + order.getId() + ".");
        return result;
    }

    private String buildString(String parameterName) {
        String result = "";
        JSObjParams params = order.getJSObjParams();
        if (params.hasParameterValue(parameterName)) {
            result = params.values().get(parameterName);
        } else
            logger.info("Parameter " + parameterName + " not set in order " + order.getId() + ".");
        return result;
    }

    @Override
    public String getInternalId() {
        return id;
    }

    @Override
    public String getName() {
        return id;
    }

    @Override
    public String getUniqueName() {
        return id;
    }

    @Override
    public List<String> getSuccessors() {
        return successors;
    }

    @Override
    public boolean hasSuccessors() {
        return (!successors.isEmpty());
    }

    @Override
    public boolean isBootstrap() {
        return isBootstrapNode;
    }

    @Override
    public JobnetNode getJobnetNodeObject() {
        return netNode;
    }

    @Override
    public String getTitle() {
        return order.getTitle();
    }

    @Override
    public JobnetNodeType getType() {
        return type;
    }

    public String getScriptToExecute() {
        return scriptToExecute;
    }

    public String getRestartOption() {
        return restartOption;
    }

    public String getEventOnError() {
        return eventOnError;
    }

    public String getEventOnSuccess() {
        return eventOnSuccess;
    }

    public boolean isErrorHandler() {
        return errorHandler;
    }

    public String getSubnet() {
        return subnet;
    }

    public String getConnector() {
        return connector;
    }
}
