package com.sos.jobnet;

import com.google.common.collect.ImmutableMap;
import com.google.common.io.Resources;
import com.sos.jobnet.classes.JobNetException;
import com.sos.testframework.h2.H2InMemoryConnection;
import com.sos.testframework.h2.H2TemporaryConnection;
import com.sos.testframework.h2.ResourceList;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.Map;
import java.util.Properties;

/**
 * In general Tests should be independent from the build machine. Therefore the test use an in memory H2 database for the test. 
 * It is also possible to use an individual hibnernate configuration. In this case you have to place your individual hinernate configuration
 * in a file named hibernate.cfg.xml in the project folder.
 *
 * @author ss
 *
 */
public class JobNetHibernateBridge {

    public final static String ssLocalHibernateConfigOracle = "C:/sos/devel/scheduler/ncfast/com.sos.jobnet/schaedi.hibernate-oracle.cfg.xml";
    public final static String ssLocalHibernateConfigH2 = "C:/sos/devel/scheduler/ncfast/com.sos.jobnet/schaedi.hibernate-h2.cfg.xml";

    private final static Logger logger = LoggerFactory.getLogger(JobNetHibernateBridge.class);

    private final static String propertyFilename = "jobnet-test.properties";
    private final static String hibernateProperty = ".hibernateConfigurationFile";
    private final static String userProperty = System.getProperty("user.name") + hibernateProperty;
    private final static String defaultProperty = "default" + hibernateProperty;

    private final File hibernateConfigurationFile;

    private final static Map<String,String> connectionMap = ImmutableMap.<String,String>builder()
            .put("com.sos.jobnet.db.EventsDBItem", "scheduler_events.sql")
            .put("com.sos.jobnet.db.JobNetNodeDBItem", "scheduler_job_net_nodes.sql")
            .put("com.sos.jobnet.db.JobNetPlanDBItem", "scheduler_job_net_plan.sql")
            .put("com.sos.jobnet.db.JobNetEdgesDBItem", "scheduler_job_net_edges.sql")
            .put("com.sos.jobnet.db.JobNetCmdHistoryDBItem", "scheduler_job_net_cmd_history.sql")
            .put("com.sos.jobnet.db.JobNetHistoryDBItem", "scheduler_job_net_history.sql")
            .put("com.sos.jobnet.db.JobNetDBItem", "scheduler_job_net.sql")
            .build();
    private H2InMemoryConnection inMemoryConnection = null;
    private H2TemporaryConnection temporaryConnection = null;

    public JobNetHibernateBridge() {

        // :h2:mem; :h2:temp
        Properties p = loadPropertiesFile();
        String hibernateConfiguration = (String) p.getProperty(userProperty);
        if (hibernateConfiguration == null)
            hibernateConfiguration = (String) p.get(defaultProperty);
        if (hibernateConfiguration == null)
            throw new RuntimeException("Neither the property " + userProperty + " nor the property " + defaultProperty + " is defined in the property file " + propertyFilename);

        File result = null;
        if(hibernateConfiguration.startsWith(":")) {
            logger.warn(hibernateConfiguration + " not implemented yet.");
            /*
            if(hibernateConfiguration.equals(":h2:mem")) {
                result = setH2InMemoryHibernateConfigurationFile();
            } else {
                if(hibernateConfiguration.equals(":h2:temp")) {
                    result = setH2TemporaryHibernateConfigurationFile();
                } else
                    throw new RuntimeException("The getExitCode " + hibernateConfiguration + " is not valid - see file " + propertyFilename);
            }
            */
        } else {
            result = setHibernateConfigurationFile(hibernateConfiguration);
        }

        this.hibernateConfigurationFile = result;
        logger.info("HibernateConfigurationFile " + hibernateConfigurationFile.getAbsolutePath() + " is used.");
    }

    private Properties loadPropertiesFile() {
        Properties p = new Properties();
        try {
            URL pr = Resources.getResource(propertyFilename);
            String prt = Resources.toString(pr, Charset.defaultCharset());
            p.load(new StringReader(prt));
        } catch (IOException e) {
            throw new RuntimeException("Error reading properties file " + propertyFilename );
        }
        return p;
    }

    /**
     * Gets a hibernate configuration. If it is not specified with setHibernateConfigurationFile() a H2 in memory
     * configuration is used.
     *
     * @return
     */
    public File getHibernateConfigurationFile() {
        return hibernateConfigurationFile;
    }

    public void close() {
        if (inMemoryConnection != null) inMemoryConnection.close();
        if (temporaryConnection != null) temporaryConnection.close();
    }

    /**
     * To use a specific (local) hibernate configuration
     * @param fileName
     */
    private File setHibernateConfigurationFile(String fileName) {
        File f = new File(fileName);
        if (!f.isFile())
            throw new JobNetException("The hibernate configuration file " + f.getAbsolutePath() + " does not exist.");
        return f;
    }

    private File setH2InMemoryHibernateConfigurationFile() {
        inMemoryConnection = new H2InMemoryConnection( new ResourceList(connectionMap) );
        inMemoryConnection.connect();
        return inMemoryConnection.createTemporaryHibernateConfiguration();
    }

    private File setH2TemporaryHibernateConfigurationFile() {
        temporaryConnection = new H2TemporaryConnection( new ResourceList(connectionMap) );
        temporaryConnection.connect();
        return temporaryConnection.createTemporaryHibernateConfiguration();
    }

}
