package com.sos.jobnet.creator;

import com.sos.JSHelper.Basics.JSToolBox;
import com.sos.jobnet.classes.JobNetConstants;
import com.sos.jobnet.classes.JobNetException;
import com.sos.jobnet.context.ContextChecker;
import com.sos.jobnet.options.JobNetCreatorOptions;
import com.sos.jobnet.options.StartModes;
import com.sos.jobnetmodel.objects.JobnetNodeType;
import com.sos.localization.Messages;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Iterator;
import java.util.List;

public class JobNetBootstrapOrder extends JobNetOrder {
	
	private final static Logger logger = LoggerFactory.getLogger(JobNetBootstrapOrder.class);
	
	private static final Messages msg = new JSToolBox(JobNetConstants.strBundleBaseName).getMessageObject();
	private final static String JOBNETC_I_0004 = "JOBNETC_I_0004"; // %1$s is successor of %2$s

	private static final String JOBNETC_E_0001 = "JOBNETC_E_0001"; 		// %1$s is successor of %2$s but %3$s is not predecessor of %4$s
	private final static String JOBNETC_E_0002 = "JOBNETC_E_0002"; 		// %1$s is predecessor of %2$s but %3$s is not successor of %4$s.
	
	@SuppressWarnings("unused")
	private static final String conSVNVersion = "$Id: JobNetBootstrapOrder.java 21146 2013-09-24 11:29:56Z ss $";

	private final StartModes startOption;
    private final ContextChecker checker;

	private static JobNetNodeList orderNodes;

	public JobNetBootstrapOrder(JobNetCreatorOptions options) {
		super(options.getOrderConnector(), options.getOrderConnector().getRootDirectory(), options.SCHEDULER_ID.Value(), options.getOrderIdWithPath(), options.getJobChainNameWithoutPath());
        logger.info("The root directory of the orders is " + getNameComponents().getFullPath());
		this.startOption = options.StartOption.asEnum();
        this.checker = new ContextChecker(getContext(), getNodeType());
        initNodeList(getMaxNodes(options));
	}
	
	private static int getMaxNodes(JobNetCreatorOptions options) {
		return Integer.parseInt(options.MaxJobNetNodes.Value());
	}
	
	private static JobNetNodeList initNodeList(int maxNodes) {
		orderNodes = new JobNetNodeList(maxNodes);
		return orderNodes;
	}

    /*
     * Set subnet and connector objects. If subnet and/or connector is not set, the bootstrap order will be subnet/connector.
     * This part has be done after the records are present in the database.
     */
    private void setSubNetAndConnectorReferences() {
        logger.info("Build the object references to the subnet and the connector nodes.");
        for(JobNetOrder o : getJobNet().values()) {

            // reference to subnet node
            {
                String subnet = o.getContext().getSubnet();
                JobNetOrder order = getNodeByNameOrNull(subnet);
                o.setSubnet( (order == null) ? this : order );
            }

            // reference to connector node
            {
                String connector = o.getContext().getConnector();
                JobNetOrder order = getNodeByNameOrNull(connector);
                o.setConnector( (order == null) ? this : order);
            }

            logger.info("JobnetNode " + o.getJobNetOrderName() + ": Subnet=" + o.getSubnet().getJobNetOrderName() + ", Connector=" + o.getConnector().getJobNetOrderName() );
        }
    }
	
	private JobNetOrder getNodeByNameOrNull(String uniqueNodeName) {
		JobNetOrder result = null;
		for(JobNetOrder o : getJobNet().values()) {
			if(o.getJobNetOrderName().equalsIgnoreCase(uniqueNodeName)) {
				result = o;
				break;
			}
		}
		return result;
	}

    /**
     * Starts the iteration of the jobnet. Beginning with the bootstrap order (this) all orders of the jobnet will be reaad from the filesystem.
     */
    public void createJobNet() {
        registerBootstrapOrder(this);
        logger.info("Iterate the jobnet based on the order files.");
        iterateJobNet(this);
        logger.info("The jobnet contains " + getJobNet().size() + " nodes after iteration.");
        setSubNetAndConnectorReferences();
    }

    /**
     * Iterates the jobnet. If StartOption requires a specific context iteration will stopped at the first node outside the context.
     * @param order
     */
    private void iterateJobNet(JobNetOrder order) {
        orderNodes.addOrderToNodesCollection(order);
        if (order.hasSuccessors()) {
            List<JobNetOrder> successors = order.getSuccessors();
            Iterator<JobNetOrder> it = successors.iterator();
            while(it.hasNext()) {
                JobNetOrder successor = it.next();
                if(!orderNodes.containsKey(successor.getJobNetNodeName())) {
                    if(isValidContext(order) && !isValidContext(successor) && successor.getNodeType() == JobnetNodeType.CONNECTOR) {
                        logger.info("The iteration of the jobnet ends with order " + successor.getOrderWithoutPath());
                        break;
                    }
                    logger.debug(msg.getMsg(JOBNETC_I_0004, successor.getJobNetNodeName(), order.getJobNetNodeName()));
                    iterateJobNet(successor);
                }
            }
        }
	}

	/**
	 * The context is valid either when the JobNetCreator starts with an option that does not perform context checking
	 * or when the given order is in the same context like the bootstrap order.
	 * 
	 * @param order
	 * @return
     */
	private boolean isValidContext(JobNetOrder order) {
		return (!startOption.checkContext() || checker.isInContext(order.getContext()));
	}

    /**
     * @return A list of all nodes in the jobnet
     */
    public JobNetNodeList getJobNet() {
        return orderNodes;
    }

    /**
     * Method to perform if the option CheckJobNetTexture is set (the jobnet has to be configured with successors AND predecessors)
     */
    public void validateJobNet() {
        for(JobNetOrder order : getJobNet().values()) {
            for(JobNetOrder successor : order.getSuccessors()) {
                if (!successor.hasPredecessor(order)) {
                    String msgText = msg.getMsg(JOBNETC_E_0001,successor.getJobNetNodeName(),order.getJobNetNodeName(),order.getJobNetNodeName(),successor.getJobNetNodeName());
                    logger.error(msgText);
                    throw new JobNetException(msgText);
                }
            }
            for(JobNetOrder predecessor : order.getPredecessors()) {
                if (!predecessor.hasSuccessor(order)) {
                    String msgText = msg.getMsg(JOBNETC_E_0002,predecessor.getJobNetNodeName(),order.getJobNetNodeName(),order.getJobNetNodeName(),predecessor.getJobNetNodeName());
                    logger.error(msgText);
                    throw new JobNetException(msgText);
                }
            }
        }
    }

}
