package com.sos.jobnet.creator;

import com.sos.JSHelper.Basics.JSToolBox;
import com.sos.JSHelper.Exceptions.JobSchedulerException;
import com.sos.jobnet.classes.*;
import com.sos.jobnet.context.ContextList;
import com.sos.jobnet.db.*;
import com.sos.jobnet.graph.JobNetGraphBuilder;
import com.sos.jobnet.graph.JobNetGraphBuilderOptions;
import com.sos.jobnet.interfaces.IJobSchedulerAdapter;
import com.sos.jobnet.options.JobNetCreatorOptions;
import com.sos.jobnet.options.SOSOptionStartJobnet;
import com.sos.jobnet.options.StartBehaviour;
import com.sos.jobnet.options.StartModes;
import com.sos.jobnetmodel.objects.JobnetNodeType;
import com.sos.jobnetmodel.objects.NodeStatus;
import com.sos.localization.Messages;
import com.sos.scheduler.model.SchedulerObjectFactory;
import com.sos.scheduler.model.objects.JSObjRunTime;
import com.sos.scheduler.model.objects.Param;
import com.sos.scheduler.model.objects.Params;
import org.joda.time.DateTime;
import org.joda.time.Interval;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import sos.spooler.Spooler;

import java.io.File;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

/**
 * \file JobNetPlanCreator.java
 * \brief creating a jobnet plan
 *  
 * \class JobNetPlanCreator
 * \brief creating a jobnet plan
 * 
 * \details
 * This class assumes a complete jobnet stored in the database for the bootstrap order which the class is
 * called for (see com.sos.jobnet.creator.JobNetCreator). Therefore the first thing calling the run method
 * is to execute the creation of the jobnet with the run method of com.sos.jobnet.creator.JobNetCreator.
 *
 * Based on this jobnet and the order definitions in the JobScheduler configuration a jobnet plan is created and
 * stored in the database in table SCHEDULER_JOB_NET_PLAN. Among other things the jobnet plan contains the
 * complete configuration of the order. This is the base to execute the jobnet.
 * \code
  \endcode
 *
 * \author Stefan Schaedlich
 * \version 1.0 - 03.04.2012 10:12:21
 */
public class JobNetPlanCreator extends JobNetCreator implements Runnable, IJobSchedulerAdapter, IDBConnection {
	
	@SuppressWarnings("unused")
	private static final String conSVNVersion = "$Id: JobNetPlanCreator.java 23331 2014-03-25 09:48:33Z sos $";

	private final static Logger logger = LoggerFactory.getLogger(JobNetPlanCreator.class);

    private final static String START_NOW = "now";
	private final static DateTimeFormatter fmtDateTime = DateTimeFormat.forPattern("yyyy-MM-dd HH:mm:ss");
//	private final static DateTime now = new DateTime();
//	private final static DateTime onDemand = now.minusMillis(now.getMillisOfDay());

	private static final Messages msg = new JSToolBox(JobNetConstants.strBundleBaseName).getMessageObject();
	private final static String JOBNETC_I_0010 = "JOBNETC_I_0010"; // job net plan created for order %1$s.
	private final static String JOBNETC_I_0012 = "JOBNETC_I_0012"; // creating jobnet with uuid %1$s.
//	private final static String JOBNETC_W_0001 = "JOBNETC_W_0001"; // the period %1$s does not represent a single start.
//	private final static String JOBNETC_E_0006 = "JOBNETC_E_0006"; // the bootstrap order %1$s contains no single_start.
	private final static String JOBNETC_E_0007 = "JOBNETC_E_0007"; // starttime %1$s is not parsable.
	private final static String JOBNETC_E_0008 = "JOBNETC_E_0008"; // the external uuid already exists.
	private final static String JOBNETC_E_0009 = "JOBNETC_E_0009"; // more than one starttime for an external uuid.
	private final static String JOBNETC_E_0015 = "JOBNETC_E_0015"; // error creating the jobnet plan
	private final static String JOBNETC_E_0018 = "JOBNETC_E_0018"; // error creating the jobnet graph

	private final Interval interval4StartTimes;
	private final boolean checkFrequency;
	private final String jobChainWithoutPath;
	private final String orderId;
	private String scriptRoot;
	// private final String externalUUID;
	private final FrequencyChecker checker;
	private final String schedulerId;
	private final String jobnetId;
	private final String jobnetParameterName;
    private final SOSOptionStartJobnet startOption;
    private final File configFile;
	private final JobNetNodeDBLayer nodeDbLayer;
	private final JobNetHistoryDBLayer historyDbLayer;
	private final JobNetPlanDBLayer planDbLayer;
	private final ConditionsDBLayer conditionsDbLayer;
	private final EventsDBLayer eventsDbLayer;
	// private final JobNetCmdHistoryDBLayer cmdHistoryDbLayer;
	private final boolean createGraph;
	private final SchedulerObjectFactory liveObjectFactory;
    private final StartModes startMode;
    private final StartBehaviour startBehaviour;

    private ContextList contextList = null;
	private List<String> collectedUUID = new ArrayList<String>();
    private JobNetHistoryDBItem historyDBItem;
    private JobNetPlanDBItem bootDBItem = null;
    private String jobnetIdQualifier = "";

    private IJobSchedulerAdapter jsAdapter = null;

	public JobNetPlanCreator(JobNetCreatorOptions options) {
		super(options);
		DateTime from = new DateTime();
		DateTime to = new DateTime(options.TimeHorizon.getEndFromNow());
		this.interval4StartTimes = new Interval(from,to);

		this.jobChainWithoutPath = options.getJobChainNameWithoutPath();
		this.orderId = options.getOrderIdWithPath();
		this.checkFrequency = options.CheckFrequency.value();
		this.jobnetId = options.JobNetId.Value();
		this.jobnetParameterName = options.JobNetIdParameterName.Value();
		this.checker = new FrequencyChecker(options);
        this.schedulerId = getOptions().SCHEDULER_ID.Value();
		// this.externalUUID = (jobnetId.isEmpty()) ? "" : jobChainWithoutPath + JobNetConstants.JOBNET_ID_DELIMITER + jobnetId;
		this.createGraph = options.CreateGraph.value();
		this.configFile = new File(getOptions().hibernate_connection_config_file.Value());
		this.planDbLayer = options.getDbConfigurationLayer().getDbLayerPlan();
        this.conditionsDbLayer = options.getDbConfigurationLayer().getDbLayerConditions();
		this.historyDbLayer = options.getDbConfigurationLayer().getDbLayerHistory();
		this.nodeDbLayer = options.getDbConfigurationLayer().getDbLayerNodes();
		// this.cmdHistoryDbLayer = new JobNetCmdHistoryDBLayer(configFile);
		this.eventsDbLayer = options.getDbConfigurationLayer().getDbLayerEvents();
		this.liveObjectFactory = options.getOrderConnector().getFactory();
        this.startMode = options.StartOption.asEnum();
        this.startBehaviour = options.StartBehaviour.asEnum();
        this.startOption = options.StartOption;
        logger.info("Timeinterval to search for start times: " + fmtDateTime.print(from) + " to " + fmtDateTime.print(to));

	}

	@Override
	public void run() {
		
		try {
		    runJobnet();
		} finally {
		}

	}

    public void setContext(ContextList contextList) {
        this.contextList = contextList;
    }

    public ContextList getContext() {
        return this.contextList;
    }

	private void runJobnet() {
		
//		if (restart.isAlways())
//			prepareRestart();		// delete all records in jobnet_plan

        if(getStartOption().newInstance()) {
            createJobnetInstance();
            if (checkFrequency)
                for (JobNetPlanDBItem planItem : getPlannedOrders())
                    checker.processJobNet(planItem);
            else
                logger.info("FrequenyChecker not executed because option CheckFrequency is set to " + getOptions().CheckFrequency.Value() + ".");
        } else
            logger.info("For " + getStartOption().name() + " it is not necessary to create a new jobnet instance in the datatabase.");

        if(contextList.isEmpty())
            throw new JobNetException("No orders stored in the jobnet model - probably setPlannedOrders() not executed first.");

		if (createGraph) {
			JobNetGraphBuilderOptions graphOptions;
			try {
				graphOptions = new JobNetGraphBuilderOptions(getOptions().Settings());
				for(String uuid : collectedUUID) {
					graphOptions.uuid_jobnet_identifier.Value(uuid);
					graphOptions.OrderId.Value(orderId);
					logger.info("graph will be created for uuid " + uuid);
					JobNetGraphBuilder graph = new JobNetGraphBuilder(graphOptions);
					graph.setBuildStatic(true);
					graph.run();
				}
			} catch (Exception e) {
				String msgText = msg.getMsg(JOBNETC_E_0018);
				logger.error(msgText);
				throw new JobSchedulerException(msgText,e);
			}
		} else {
            logger.info("GraphBuilder not executed because option CreateGraph is set to " + getOptions().CreateGraph.Value() + ".");
        }

	}

	private void createJobnetInstance() {

        List<JobNetPlanDBItem> plannedOrders = new ArrayList<JobNetPlanDBItem>();

        logger.info("Create a new jobnet instance.");
        try {
			super.run(); // ensures the representation of the jobNet table JOB_NET_NODES and JOB_NET_EDGES of the database
	
			/*
			 * In case of an external given uuid it has to be unique.
			if (hasExternalUUID()) {
				logger.info("Creating jobnnet instance with external id " + externalUUID);
				if (isJobNetInstance(planDbLayer, externalUUID)) {
					String msgText = msg.getMsg(JOBNETC_E_0008, externalUUID);
					logger.error(msgText);
					throw new JobSchedulerException(msgText);
				}
			}
			 */

			JobNetBootstrapOrder bootstrapOrder = getBootstrapOrder();
            historyDbLayer.beginTransaction();
            planDbLayer.beginTransaction();

			JSObjRunTime runTime = bootstrapOrder.getJSObjRunTime();
			if (bootstrapOrder.isOnDemandOrder() || startBehaviour == StartBehaviour.on_demand) {
				bootDBItem = createJobNetPlan4DemandOrderInDB(createUUID());
				logger.info("Order " + bootstrapOrder.getJobNetNodeName() + " will be start on demand");
				plannedOrders.add(bootDBItem);
			} else {
				logger.debug("hasPeriod? " + runTime.hasPeriod());
				logger.debug("hasAt? " + runTime.hasAt());
				logger.debug("hasDate? " + runTime.hasDate());
				logger.debug("hasWeekdays? " + runTime.hasWeekdays());
				logger.debug("hasMonth? " + runTime.hasMonth());
				logger.debug("hasMonthdays? " + runTime.hasMonthdays());
				logger.debug("hasUltimos? " + runTime.hasUltimos());
				if (runTime.hasSubsequentRunTimes()) {
					
					/*
					 * If the bootstrap order contains more than one start time, it is not possible to set the uuid external.
					 */
					List<DateTime> startTimes = runTime.getDtSingleStarts(interval4StartTimes);
                    if (startTimes.size() == 0) {
                        logger.info("No single starts found in runtime definition of order " + bootstrapOrder.getJobNetOrderName() + " - Order will be start immediately.");
                        bootDBItem = createJobNetPlan4OrderInDB(createUUID(), START_NOW, 0);
                        logger.info("Order " + bootstrapOrder.getJobNetNodeName() + " will be start immediately");
                        plannedOrders.add(bootDBItem);
                    } else {
                        logger.info(startTimes.size() + " runtimes calculated for " + bootstrapOrder.getJobNetNodeName());
                        for (DateTime startTime : startTimes) {
                            String startString = fmtDateTime.print(startTime);
                            logger.info("Start time at " + startString);
                        }

                        int cnt = 0;
                        for (DateTime startTime : startTimes) {
                            String startString = fmtDateTime.print(startTime);
                            bootDBItem = createJobNetPlan4OrderInDB(createUUID(), startString, cnt);
                            logger.info("Order " + bootstrapOrder.getJobNetNodeName() + " will be start at " + startString);
                            plannedOrders.add(bootDBItem);
                            cnt++;
                        }
                    }

				} else {
					bootDBItem = createJobNetPlan4OrderInDB(createUUID(), START_NOW, 0);
					logger.info("Order " + bootstrapOrder.getJobNetNodeName() + " will be start immediately");
					plannedOrders.add(bootDBItem);
				}
			}

            // planDbLayer.commit();
            historyDbLayer.commit();

            logger.info("Try to build context list for order " + bootDBItem.getJobnetNodeDBItem().getNode());
            this.contextList = new ContextList(planDbLayer,plannedOrders,bootDBItem);

        } catch(Exception e) {
            throw new JobNetException("Error creating jobnet plan.",e);

		} finally {
        }

	}

    /**
     * Creates a unique identifier for the jobnet instance. If no JobNetId id is given via option it will be a system generated value.
     *
     * Otherwise a key build of jobchain.jobnetid take place as id. If it is not unique it will be expand with a number that makes it unique.
     * Independent jobnet instances are indicated with "_ind" in any case.
     *
     * @return
     */
    private String createUUID() {
        String uuid = null;
        if(jobnetId.isEmpty()) {
            uuid = UUID.randomUUID().toString();
        }  else {
            String baseId = jobChainWithoutPath + JobNetDAOConstants.JOBNET_ID_DELIMITER + jobnetId;
            int subId = 0;
            String independent = (startMode == StartModes.start_independent) ? JobNetConstants.JOBNETID_QUALIFIER : "";
            uuid = baseId + independent;
            jobnetIdQualifier = independent;
            while (historyDbLayer.isJobNetInstance(uuid)) {
                logger.info("The uuid " + subId + " already exists, create a new one ...");
                subId++;
                jobnetIdQualifier = independent + "_" + subId;
                uuid = baseId + jobnetIdQualifier;
            }
        }
        logger.info("The uuid for the jobnet is " + uuid);
        return uuid;
    }

	private JobNetPlanDBItem createJobNetPlan4DemandOrderInDB(String uuid) {
		String orderId = getUniqueOrderId(0);
		return createJobNetPlanAndWrite2Db(uuid, START_NOW, true, orderId);
	}
	
	private JobNetPlanDBItem createJobNetPlan4OrderInDB(String uuid, String startTime, int orderIndex) {
		String orderId = getUniqueOrderId(orderIndex);
		return createJobNetPlanAndWrite2Db(uuid, startTime, false, orderId);
	}

    private JobNetHistoryDBItem createJobNetHistoryEntryAndWrite2Db(String uuid, Integer numberOfNodes) {
        JobNetHistoryDBItem historyItem = new JobNetHistoryDBItem();
        historyItem.setSchedulerId(schedulerId);
        historyItem.setNetId(getNetId());
        historyItem.setUuid(uuid);
        historyItem.setStartNodeId(getBootstrapOrder().getNodeDBItem().getNodeId());
        historyItem.setStatus(JobNetStatus.NOT_PROCESSED.getIndex());
        historyItem.setNumberOfFinishedNodes(0);
        historyItem.setNumberOfNodes( numberOfNodes);
        historyItem.setExitCode(0);
        historyItem.setExitMessage("");
        historyItem.setJobNetId(jobnetId);
        historyItem.setJobNetIdParameterName(jobnetParameterName);
        historyItem.setCreated(new DateTime());
        historyItem.setCreatedBy(this.getClass().getName());
        historyDbLayer.save(historyItem);
        return historyItem;
    }

	private JobNetPlanDBItem createJobNetPlanAndWrite2Db(String uuid, String startTime, boolean isOnDemand, String orderId) {

		JobNetPlanDBItem startNode = null;
		logger.info(msg.getMsg(JOBNETC_I_0012,uuid));

        getOptions().uuid_jobnet_identifier.Value(uuid);

        Integer numberOfNodes = getBootstrapOrder().getJobNet().size();
        this.historyDBItem = createJobNetHistoryEntryAndWrite2Db(uuid, numberOfNodes);

        List<JobNetPlanDBItem> newRecords = new ArrayList<JobNetPlanDBItem>();
        String bootstrapOrderId = getBootstrapOrder().getOrderWithoutPath();
		for (JobNetOrder order : getBootstrapOrder().getJobNet().values()) {

            boolean bootstrapOrder = bootstrapOrderId.equals(order.getOrderWithoutPath());

			order.setId(order.getId());
			JobNetPlanDBItem item = new JobNetPlanDBItem();
			item.setUuid(uuid);                                             // a unique id for the jobnet
            item.setHistoryId(historyDBItem.getHistoryId());                // a reference to the jobnet header in SCHEDULER_JOB_NET_HISTORY

            // Save the global jobnet parameters in a separate database field
            item.setParamXml(getGlobalJobnetParameters(uuid).toXMLString());

            // Save the order xml (without params)
            item.setOrderXml(order.toXMLString());
            item.setJobnetNodeDBItem(order.getNodeDBItem());

            item.setNodeType(order.getNodeType().name());
            item.setNodeId(order.getNodeDBItem().getNodeId());
            item.setSubnetId(order.getSubnet().getNodeDBItem().getNodeId());
            item.setConnectorId(order.getConnector().getNodeDBItem().getNodeId());
            item.setScriptToExecute(order.getScriptToExecute());
            item.setUseErrorHandler(order.getUseErrorHandler());

            if(bootstrapOrder) {
                item.setParentId(0L);
            } else {
                if(order.getNodeType() == JobnetNodeType.CONNECTOR)
                    item.setParentId(order.getSubnet().getNodeDBItem().getNodeId());
                if(order.getNodeType() == JobnetNodeType.SUBNET)
                    item.setParentId(order.getConnector().getNodeDBItem().getNodeId());
                if(order.getNodeType() == JobnetNodeType.JOB)
                    item.setParentId(order.getConnector().getNodeDBItem().getNodeId());
            }

            item.setNodeName(order.getNodeDBItem().getNode());

            // runtime informations
            Integer NOT_PROCESSED = NodeStatus.NOT_PROCESSED.getIndex();
			item.setStatusNode(NOT_PROCESSED);
			item.setStatusWaiter(NOT_PROCESSED);
			item.setStatusRunner(NOT_PROCESSED);
			item.setStatusDispatcher(NOT_PROCESSED);
            item.setExitCode(0L);
            item.setTaskId(0L);
            item.setFrequencyOverwrite(false);
            item.setIsWaiterSkipped(false);

            // only for job nodes the runner will be executed
            // item.setIsRunnerSkipped( shouldRunnerSkipped(order) );
            item.setIsRunnerSkipped(false);
            item.setIsDispatcherSkipped(false);
            item.setIsRunnerOnDemand(isOnDemand);		// is set by FrequenyChecker (except for bootstarp order)
            //item.setBootstrap(order.isBootstrap());
            item.setBootstrap(bootstrapOrder);

            // record history
			item.setCreated(new DateTime());
			item.setCreatedBy(JobNetPlanCreator.class.getName());
			item.setModified(new DateTime());
			item.setModifiedBy(JobNetDAOConstants.CURRENT_USER);

			//if (order.isBootstrap()) {
			if (bootstrapOrder) {
				try {
					item.setPlannedStartime(startTime);         // maybe it is "now" - this causes an error
				} catch (ParseException e) {
					String msgText = msg.getMsg(JOBNETC_E_0007,order.getAt());
					logger.error(msgText,e);
					throw new JobNetException(msgText);
				}
			}

			planDbLayer.save(item);
            newRecords.add(item);

            createConditions(order,item);

            //if (order.isBootstrap())
            if (bootstrapOrder)
                startNode = item;

		}

        logger.info(msg.getMsg(JOBNETC_I_0010,startNode.getJobnetNodeItem().getNode()));
        collectedUUID.add(uuid);
		return startNode;
	}

    private void createConditions(JobNetOrder forOrder, JobNetPlanDBItem item) {
        if(forOrder.isOnDemandOrder())
            conditionsDbLayer.createOnDemandCondition(item);

        ParameterSubstitution s = new ParameterSubstitution(
                getOptions().uuid_jobnet_identifier.Value(),
                getOptions().getJobChainNameWithoutPath(),
                getOptions().JobNetIdParameterName.Value()
        );
        String eventString = forOrder.getWaitForEvents().trim();
        try {
            if(!eventString.isEmpty()) {
                String[] eventList = forOrder.getWaitForEvents().split(JobNetDAOConstants.JOBNET_PARAMETER_COLLECTION_DELIMITER);
                for(String eventName : eventList) {
                    logger.info("Proceeding event " + eventName);
                    String eventValue = s.replace(eventName);
                    conditionsDbLayer.createSingleEventCondition(item,eventValue);
                }
            }
        } catch(Exception e) {
            throw new RuntimeException("Error creating event conditions (eventstring=" + eventString + ").",e);
        }

        for(JobNetOrder p : forOrder.getPredecessors()) {
            conditionsDbLayer.createPredecessorCondition(item, p.getOrderWithoutPath() + JobNetDAOConstants.JOBNET_ID_DELIMITER + this.jobChainWithoutPath + JobNetDAOConstants.JOBNET_ID_DELIMITER + jobnetId);
        }
    }

	/**
	 * Creates a Params collection with all global jobnet parameters.
	 */
	private Params getGlobalJobnetParameters(String uuid) {

        Params params = liveObjectFactory.createParams();
		String paramName = getOptions().JobNetIdParameterName.Value();
		if(!paramName.isEmpty()) {
            String jobNetIdParameterName = getOptions().JobNetIdParameterName.Value();
            addParam(params, JobNetGlobalParams.JobnetIdParameterName.name(), jobNetIdParameterName );
            addParam(params, jobNetIdParameterName, jobnetId);
		}

        addParam(params, JobNetGlobalParams.uuid_jobnet_identifier.name(), uuid);
        addParam(params, JobNetGlobalParams.JobNetId.name(), jobnetId);
        addParam(params, JobNetGlobalParams.JobNetIdQualifier.name(), jobnetIdQualifier);
        addParam(params, JobNetGlobalParams.JobNet.name(), getBootstrapOrder().getJobNetJobChainNameWithPath());
        addParam(params, JobNetGlobalParams.JobChainName.name(), jobChainWithoutPath);
        addParam(params, JobNetGlobalParams.ScriptRoot.name(), scriptRoot);
        addParam(params, JobNetGlobalParams.StartOption.name(), startOption.asEnum().name());

		return params;
	}

    private void addParam(Params addToParams, String name, String value) {
        Param p = liveObjectFactory.createParam();
        p.setName(name);
        p.setValue(value);
        addToParams.getParamOrCopyParamsOrInclude().add(p);
    }
	
	private String getUniqueOrderId(int orderIndex) {
		String result = getBootstrapOrder().getJobNetOrderName();
		return (orderIndex > 0) ? result + "_" + orderIndex : result;
	}

    public JobNetHistoryDBItem getHistoryDBItem() {
        return historyDBItem;
    }

    public JobNetPlanDBItem getStartNode() {
        return bootDBItem;
    }

	public List<JobNetPlanDBItem> getPlannedOrders() {
		return contextList.getNodesInContext();
	}

	public void registerJSAdapter(IJobSchedulerAdapter adapter) {
        logger.info("Order Starter " + adapter.getClass().getName() + " registered.");
		this.jsAdapter = adapter;
        this.scriptRoot = jsAdapter.getSpooler().variables().value(JobNetConstants.SCRIPT_ROOT_GLOBAL_PARAMETER_NAME);
        if(scriptRoot == null)
            throw new JobNetException("Parameter 'ScriptRoot' not configured in scheduler.xml.");
	}

    public IJobSchedulerAdapter getJSAdapter() {
        return jsAdapter;
    }

    @Override
    public void setOrderStateText(String text) {
        if (jsAdapter != null)
            jsAdapter.setOrderStateText(text);
        else
            logger.warn("No object of type IJobSchedulerAdapter is registered to set the order state text.");
    }

    @Override
    public Spooler getSpooler() {
        Spooler result = null;
        if (jsAdapter != null)
            result = jsAdapter.getSpooler();
        else
            logger.warn("No object of type IJobSchedulerAdapter is registered to start an order.");
        return result;
    }

    @Override
	public void startOrder(JobNetPlanDBItem item, StartBehaviour startBehaviour) {
		if (jsAdapter != null)
			jsAdapter.startOrder(item, startBehaviour);
        else
			logger.warn("No object of type IJobSchedulerAdapter is registered to start an order.");
	}

    @Override
    public void resumeOrder() {
        //TODO To change body of implemented methods use File | Settings | File Templates.
    }

    public List<String> getUUIDs() {
        return collectedUUID;
    }

}
