package com.sos.jobnet.creator;

import com.sos.JSHelper.Basics.JSToolBox;
import com.sos.jobnet.classes.JobNetConstants;
import com.sos.jobnet.classes.JobNetException;
import com.sos.jobnet.db.*;
import com.sos.jobnet.options.JobNetCreatorOptions;
import com.sos.jobnet.options.StartBehaviour;
import com.sos.jobnetmodel.objects.NodeStatus;
import com.sos.localization.Messages;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;

public class JobNetPlanRestarter extends JobNetPlanExecutor implements Runnable, IDBConnection {
	
	@SuppressWarnings("unused")
	private static final String conSVNVersion = "$Id$";
	
	private final static int NOT_PROCESSED = NodeStatus.NOT_PROCESSED.getIndex();

	private static final Messages msg = new JSToolBox(JobNetConstants.strBundleBaseName).getMessageObject();
	private final static String JOBNETC_E_0011 = "JOBNETC_E_0011"; // no jobnet plan found for uuid %1$s.
	private final static String JOBNETC_E_0012 = "JOBNETC_E_0012"; // jobnet plan with uuid %1$s has more than one bootstrap order.
	private final static String JOBNETC_E_0013 = "JOBNETC_E_0013"; // jobnet plan with uuid %1$s has no bootstrap order.

	private final static Logger logger = LoggerFactory.getLogger(JobNetPlanRestarter.class);
	
	private final JobNetPlanDBLayer planDbLayer;
	private final EventsDBLayer eventsDbLayer;
	private final JobNetHistoryDBLayer historyDbLayer;
	private final JobNetHistoryDBItem historyRecord;
	private final StartBehaviour startBehaviour;
	private final String uuid;
    private final JobNetConifigurationLayer dbLayer;

	private JobNetPlanDBItem bootStrapOrder = null;

	public JobNetPlanRestarter(JobNetCreatorOptions options) {
		super(options);
		this.uuid = options.uuid_jobnet_identifier.Value();
        this.dbLayer = options.getDbConfigurationLayer();
		this.planDbLayer = dbLayer.getDbLayerPlan();
		this.eventsDbLayer = dbLayer.getDbLayerEvents();
		this.historyDbLayer = dbLayer.getDbLayerHistory();
		this.historyRecord = historyDbLayer.getJobNetInstance(uuid);
		this.startBehaviour = options.StartBehaviour.asEnum();
	}

	@Override
	public void run() {

		try {
			logger.info("Try to restart jobnet with id " + uuid);
			if(historyDbLayer.isJobnetRunning(historyRecord)) {
				String msg = "Jobnet instance " + uuid + " is already running - could not restart.";
				logger.error(msg);
				throw new JobNetException(msg);
			}
			
			JobNetPlanDBItem bootStrapOrder = prepareForRestart();
			super.startOrder(bootStrapOrder,startBehaviour);
		} catch(RuntimeException e) {
			throw e;
		} finally {
		}
	}
	
	private JobNetPlanDBItem prepareForRestart() {
		
		try {
			logger.info("Try to restart jobnet with id " + uuid);
			List<JobNetPlanDBItem> jobNetPlan = getJobNetPlan(uuid);
			
			planDbLayer.beginTransaction();
			for(JobNetPlanDBItem item : jobNetPlan) {
				resetJobNetPlanRecord(item);
			}
			if (bootStrapOrder == null) {		// no bootstrap order
				String msgText = msg.getMsg(JOBNETC_E_0013,uuid);
				logger.error(msgText);
				throw new JobNetException(msgText);
			}
			planDbLayer.commit();

			eventsDbLayer.beginTransaction();
			eventsDbLayer.deleteEventsForClass(uuid);
			eventsDbLayer.commit();
			
			logger.info("uuid " + uuid + " prepared for restart.");
			
		} catch(RuntimeException e) {
			String msg = "Error preparing jobnet instance " + uuid + " for restart.";
			logger.error(msg,e);
			throw new JobNetException(msg,e);
		}
		return bootStrapOrder;
	}
	
	private void resetJobNetPlanRecord(JobNetPlanDBItem item) {
		
		if (item.getBootstrap() ) {		// multiple bootstrap orders
			if (bootStrapOrder != null) {
				String msgText = msg.getMsg(JOBNETC_E_0012,uuid);
				logger.error(msgText);
				throw new JobNetException(msgText);
			}
			bootStrapOrder = item;
		}
		
		item.setMessage("");
		item.setExitCode(0L);
		item.setIsDispatcherSkipped(false);
		item.setIsRunnerSkipped(false);
		item.setIsWaiterSkipped(false);
		item.setStatusNode(NOT_PROCESSED);
		item.setStatusRunner(NOT_PROCESSED);
		item.setStatusWaiter(NOT_PROCESSED);
		planDbLayer.update(item);
	}
	
	private List<JobNetPlanDBItem> getJobNetPlan(String uuid) {
		JobNetPlanFilter filter = new JobNetPlanFilter();
		filter.setUuid(uuid);
		planDbLayer.setFilter(filter);
		List<JobNetPlanDBItem> list = planDbLayer.getJobnetPlanList(0);
		if (list.size() == 0) {
			String msgText = msg.getMsg(JOBNETC_E_0011,uuid);
			logger.error(msgText);
			throw new JobNetException(msgText);
		}
		return list;
	}

    @Override
    public void closeSession() {
        dbLayer.closeSession();
    }
}
