package com.sos.jobnet.db;

import com.sos.jobnet.classes.JobNetDAOException;
import com.sos.jobnetmodel.objects.JobnetNodeType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created with IntelliJ IDEA.
 *
 * @uthor ss
 * at 05.09.13 19:48
 */
public class JobNetContextDBReader {

    private static final Logger logger = LoggerFactory.getLogger(JobNetContextDBReader.class);

    private final JobNetEdgesDBLayer edgesDBLayer;
    private final JobNetPlanDBLayer planDBLayer;
    private final JobNetPlanDBItem planDBItem;
    private final List<JobNetPlanDBItem> nodeList = new ArrayList<JobNetPlanDBItem>();
    private final JobNetPlanTree tree;

    private Map<Long,JobNetPlanDBItem> buffer = new HashMap<Long,JobNetPlanDBItem>();
    private List<JobNetPlanDBItem> result = new ArrayList<JobNetPlanDBItem>();

    protected JobNetContextDBReader(JobNetPlanDBLayer planDBLayer, JobNetPlanDBItem planDBItem) {
        if (planDBItem == null)
            throw new JobNetDAOException("The constructor parameter planDBItem should not be null");
        this.planDBLayer = planDBLayer;
        this.edgesDBLayer = new JobNetEdgesDBLayer(planDBLayer.getConfigurationFile());
        this.planDBItem = planDBItem;
        this.tree = new JobNetPlanTree(planDBItem);
        resolveJobnet();
    }

    public List<JobNetPlanDBItem> getMemberList() {
        return  nodeList;
    }

    public JobNetPlanTree getMemberTree() {
        return tree;
    }

    private void resolveJobnet() {
        addMember(planDBItem);
        iterateJobnet(planDBItem,tree);
        logger.info("The memberlist contains " + nodeList.size() + " elements.");
    }

    private void iterateJobnet(JobNetPlanDBItem item, JobNetPlanTree currentTree) {
        List<JobNetPlanDBItem> childNodes = getNodesOfNextLevel(item);
        for(JobNetPlanDBItem child : childNodes) {
            JobNetPlanTree newTree = currentTree.addSuccessor(child);
            addMember(child);
            iterateJobnet(child,newTree);
        }
    }

    private void addMember(JobNetPlanDBItem item) {
        nodeList.add(item);
    }

    public List<JobNetPlanDBItem> getNodesOfNextLevel(JobNetPlanDBItem item) {
        List<JobNetPlanDBItem> result = new ArrayList<JobNetPlanDBItem>();
        while(true) {
            if (isSubnet(item)) {
                result = getConnectors(item);
                break;
            }
            if (isConnector(item)) {
                result = getSubnetOrJob(item);
                break;
            }
            break;          // a job has no entries at next level
        }
        return result;
    }

    private List<JobNetPlanDBItem> getConnectors(JobNetPlanDBItem item) {
        if (isSubnet(item)) {
            Long subnetId = item.getNodeId();
            JobNetPlanFilter jobNetPlanFilter = new JobNetPlanFilter();
            jobNetPlanFilter.setSubnetId(subnetId);
            jobNetPlanFilter.setNodeType(JobnetNodeType.CONNECTOR.name());
            planDBLayer.setFilter(jobNetPlanFilter);
            result = planDBLayer.getJobnetPlanList(0);

            // Tranform the result in an indexed HashMap
            buffer.clear();
            for(JobNetPlanDBItem i : result) {
                buffer.put(i.getPlanId(),i);
            }
            result.clear();

            // put the nodes in the right sequence to the resultlist
            iterateSubnet(item);

        }
        return result;
    }

    private void iterateSubnet(JobNetPlanDBItem item) {
        List<JobNetEdgesDBItem> edges = edgesDBLayer.getSuccessors(item.getJobNetNodeDBItem());
        for(JobNetEdgesDBItem edge : edges) {
            Long childId = edge.getChildNodeId();
            JobNetPlanDBItem childNode = planDBLayer.getByNodeIdOrNull(childId);
            if(buffer.containsKey(childNode.getPlanId()))
                result.add(childNode);
            if(result.size() == buffer.size())
                break;
            iterateSubnet(childNode);
        }
    }

    private List<JobNetPlanDBItem> getSubnetOrJob(JobNetPlanDBItem item) {
        List<JobNetPlanDBItem> result = new ArrayList<JobNetPlanDBItem>();
        if (isConnector(item)) {

            Long connectorId = item.getNodeId();

            // get all JOBS under the CONNECTOR
            JobNetPlanFilter jobNetPlanFilter = new JobNetPlanFilter();
            jobNetPlanFilter.setConnectorId(connectorId);
            planDBLayer.setFilter(jobNetPlanFilter);
            jobNetPlanFilter.setNodeType(JobnetNodeType.JOB.name());
            result = planDBLayer.getJobnetPlanList(0);

            // get all SUBNETs under the CONNECTOR
            List<JobNetEdgesDBItem> edges = edgesDBLayer.getSuccessors(item.getJobNetNodeDBItem());
            for(JobNetEdgesDBItem edge : edges) {
                Long childId = edge.getChildNodeId();
                JobNetPlanDBItem childNode = planDBLayer.getByNodeIdOrNull(childId);
                if(isSubnet(childNode))
                    result.add(childNode);
            }
        }
        return result;
    }

    private boolean isSubnet(JobNetPlanDBItem item) {
        return item.getNodeType().equalsIgnoreCase(JobnetNodeType.SUBNET.name());
    }

    private boolean isConnector(JobNetPlanDBItem item) {
        return item.getNodeType().equalsIgnoreCase(JobnetNodeType.CONNECTOR.name());
    }

    private boolean isJob(JobNetPlanDBItem item) {
        return item.getNodeType().equalsIgnoreCase(JobnetNodeType.JOB.name());
    }

}
