package com.sos.jobnet.graph;

import com.sos.JSHelper.Basics.JSToolBox;
import com.sos.graphviz.GraphIO;
import com.sos.graphviz.enums.FileType;
import com.sos.graphviz.enums.RankDir;
import com.sos.hibernate.options.HibernateOptions;
import com.sos.jobnet.classes.JobNetConstants;
import com.sos.jobnet.classes.JobNetException;
import com.sos.jobnet.creator.JobNetPlanExecutor;
import com.sos.jobnetmodel.graph.GraphBuilder;
import com.sos.jobnetmodel.graph.GraphOptions;
import com.sos.jobnetmodel.objects.JobnetConverter;
import com.sos.jobnetmodel.objects.JobnetConverterImpl;
import com.sos.localization.Messages;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;

public class JobNetGraphBuilder implements Runnable {

	private final static Logger logger = LoggerFactory.getLogger(JobNetGraphBuilder.class);
	private static final Messages msg = new JSToolBox(JobNetConstants.strBundleBaseName).getMessageObject();

	private final static String graphSubfolder = "graphs";

	private FileType graphType = FileType.svg;
	private String extension = "." + graphType.name();

	private final String graphDir;
	private final String uuid;
	private final HibernateOptions hibernateOptions;
	private final File targetDir;
	private final String size;
	private final String ratio;
	private final String orderWithoutPath;
	File targetFile;

	private boolean buildStaticFile = false;
	private boolean isJobnetPresentInDb = false;

	public JobNetGraphBuilder(JobNetGraphBuilderOptions options) {
		String conMethodName = JobNetPlanExecutor.class.getSimpleName();
		this.hibernateOptions = options;
		this.uuid = options.uuid_jobnet_identifier.Value();
        this.orderWithoutPath = options.OrderId.Value();
        this.size = options.GraphSize.Value();
        this.ratio = options.GraphRatio.Value();
		this.graphDir = graphDir(options);
		this.graphType = graphType(options);
		this.extension = "." + graphType.name();

		this.targetDir = new File(graphDir);
		targetDir.mkdirs();
		logger.debug(msg.getMsg(JobNetConstants.JOBNET_D_0001, conMethodName, "UUID", uuid));
	}

	private String graphDir(JobNetGraphBuilderOptions options) {
		String dir = options.GraphDir.Value().trim();
		if (dir.isEmpty()) {
			dir = options.SCHEDULER_DIRECTORY.Value() + "/" + graphSubfolder;
			logger.info("Parameter 'GraphDir' not specified - using default " + dir);
		}
		logger.info("Graphs will be stored in folder " + dir);
		return dir;
	}

	private FileType graphType(JobNetGraphBuilderOptions options) {
		FileType gType = null;
		try {
			gType = FileType.fromText(options.GraphType.Value());
		}

		catch (Exception e) {
			logger.error(String.format("Wrong GraphType: %s", options.GraphType.Value()));
			e.printStackTrace();
		}
		return gType;

	}

	@Override
	public void run() {

        logger.info("Create a graph for jobnet with uuid " + uuid);
		JobnetConverter converter = convertJobnetFromDB();
		if (converter != null) {
			logger.debug("Number of connectors: " + converter.numberOfConnectores());
			logger.debug("Number of head: " + converter.numberOfHeads());
			logger.debug("Number of normal nodes: " + converter.numberOfNormalNodes());
			logger.debug("Total Number of nodes: " + converter.getNodeList().size());
			logger.debug("Number of containers: " + converter.numberOfContainer());
			createGraph(converter);
		}

	}

	private JobnetConverter convertJobnetFromDB() {
		JobnetConverter converter = null;
		try {
			JobNetPlanAdapter adapter = new JobNetPlanAdapter(hibernateOptions, uuid, orderWithoutPath);
            adapter.setAllowInvalidNodes(true);
			logger.debug("Number of adapted nodes: " + adapter.numberOfNodes());
			isJobnetPresentInDb = adapter.isJobnetPresentInDB();
			if (isJobnetPresentInDb) {
				converter = new JobnetConverterImpl(adapter);
				converter.run();
			} else {
				logger.warn("A jobnet with uuid " + uuid + " is not present in the DB.");
			}
		} catch (Exception e) {
			String msg = "Error converting jobnet from database.";
			logger.error(msg, e);
			throw new JobNetException(msg, e);
		}
		return converter;
	}

	private void createGraph(JobnetConverter converter) {
		GraphBuilder gb = null;
		try { 
			gb = new GraphBuilder(converter, new GraphOptions(RankDir.TB));
            if (!size.isEmpty()) gb.getGraphProperties().setSize(size);
            if (!ratio.isEmpty()) gb.getGraphProperties().setRatio(ratio);
 
			gb.execute();
		} catch (Exception e) {
			String msg = "Error creating graph.";
			logger.error(msg, e);
			throw new JobNetException(msg, e);
		}

		try {
			GraphIO io = new GraphIO(gb);
			File tempFile = File.createTempFile("graphviz_", ".tmp");
			targetFile = prepareTargetFile( getOutputFilename(converter.getBootstrapNodeName()) );
			logger.debug(gb.getContent());
			io.writeGraphToFile(graphType, tempFile);
			if (targetFile.exists())
                throw new RuntimeException("Targetfile exists: " + targetFile.getAbsolutePath() + " - could not move graph from " + tempFile.getAbsolutePath());
    		tempFile.renameTo(targetFile);
		} catch (IOException e) {
			String msg = "Error creating jobnet graph.";
			logger.error(msg, e);
			throw new JobNetException(msg, e);
		}

	}

    private File prepareTargetFile(String baseName) {
        File resultFile = new File(targetDir, baseName + extension);
        if (resultFile.exists()) {
            if (!resultFile.delete()) {
                logger.warn("Error deleting targetfile " + resultFile);
                int cnt = 0;
                while(resultFile.exists()) {
                    cnt++;
                    resultFile = new File(targetDir, baseName + "_" + cnt + extension);
                }
            }
        }
        logger.info("Output to file " + resultFile.getAbsolutePath());
        return resultFile;
    }

	private String getOutputFilename(String baseName) {
		return (buildStaticFile) ? baseName : uuid;
	}

	/**
	 * This flag determines if the file created contains the uuid in its name.
	 * 
	 * @param staticFile
	 */
	public void setBuildStatic(boolean staticFile) {
		this.buildStaticFile = staticFile;
	}

	public boolean isJobnetPresentInDB() {
		return isJobnetPresentInDb;
	}

    public String getExtension() {
        return extension;
    }

    public File getTargetFile() {
        return targetFile;
    }

}
