package com.sos.jobnet.graph;

import com.sos.jobnet.classes.JobNetException;
import com.sos.jobnet.db.JobNetEdgesDBItem;
import com.sos.jobnet.db.JobNetEdgesDBLayer;
import com.sos.jobnet.db.JobNetPlanDBItem;
import com.sos.jobnetmodel.live.LockUsed;
import com.sos.jobnetmodel.objects.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

/*
 * Adapter class to convert one node of the database representation of the jobnet into an internal jobnet node.
 */
public class JobNetPlanRecordAdapter implements ExternalJobnetNodeObject {
	
	private static Logger logger = LoggerFactory.getLogger(JobNetPlanRecordAdapter.class);
	
	private final JobNetEdgesDBLayer edgesDBReader;
	private final JobNetPlanDBItem dbRecord;
    private final List<String> successors;
    private final boolean isBootstrap;
    private final String uniqueName;
    private final String connector;
    private final String subnet;
    private final JobnetNode netNode;
    
    private final static Pattern nodePattern = Pattern.compile("_step\\d\\d\\d");
	
	public JobNetPlanRecordAdapter(JobNetEdgesDBLayer edgesDBReader, JobNetPlanDBItem dbRecord) {
		this.dbRecord = dbRecord;
		this.edgesDBReader = edgesDBReader;
		this.successors = readSuccessors();
		this.isBootstrap = dbRecord.getBootstrap();
		this.uniqueName = dbRecord.getJobnetNodeDBItem().getNode();
        this.connector = (dbRecord.getConnectorNodeDBItem() != null) ? dbRecord.getConnectorNodeDBItem().getJobNetNodeDBItem().getNode() : "";
        this.subnet = (dbRecord.getSubNetNodeDBItem() != null) ? dbRecord.getSubNetNodeDBItem().getJobNetNodeDBItem().getNode() : "";
        this.netNode = new JobnetNodeImpl(String.valueOf(dbRecord.getNodeId()), getType());

		logger.debug(getType().name() + ": Adapt jobnet node " + dbRecord.getJobnetNodeDBItem().getNode() + " (id " + dbRecord.getNodeId() + ")");
        // netNode.setSuccessors(successors);
        netNode.setExternalNodeObject(this);
        netNode.setName(dbRecord.getJobnetNodeDBItem().getNode());
        
        NodeStatus s = NodeStatus.valueOf(dbRecord.getStatusNode());
        netNode.setNodeStatus(s);
	}
	
	private List<String> readSuccessors() {
		List<String> result = new ArrayList<String>();
		List<JobNetEdgesDBItem> jobnetEdgesList = edgesDBReader.getSuccessors(dbRecord.getJobnetNodeDBItem());
		for(JobNetEdgesDBItem edge : jobnetEdgesList) {
			logger.debug("Successor " + edge.getChildNodeId() + " detected.");
			result.add( String.valueOf(edge.getChildNodeId()) );
		}
		return result;
	}

	@Override
	public List<String> getSuccessors() {
		return successors;
	}

    @Override
    public JobnetNodeType getType() {
        JobnetNodeType result = null;
        try {
            result = JobnetNodeType.valueOf(dbRecord.getNodeType());
        } catch(Exception e) {
            throw new JobNetException("The nodetype " + dbRecord.getNodeType() + " of order " + dbRecord.getJobnetNodeDBItem().getNode() + " for UUID " + dbRecord.getUuid() + " is not valid.");
        }
        return result;
    }

    @Override
	public boolean hasSuccessors() {
		return (!successors.isEmpty());
	}

	@Override
	public boolean isBootstrap() {
		return isBootstrap;
	}

    @Override
    public String getInternalId() {
        // return uniqueName;
        return String.valueOf(dbRecord.getNodeId());
    }

    @Override
    public String getName() {
        return dbRecord.getJobNetNodeDBItem().getNode();
    }

    @Override
	public String getUniqueName() {
		return uniqueName;
	}

    @Override
    public String getConnector() {
        return connector;
    }

    @Override
    public String getSubnet() {
        return subnet;
    }

    @Override
	public JobnetNode getJobnetNodeObject() {
		return netNode;
	}

    @Override
    public String getTitle() {
        return dbRecord.getJobNetNodeDBItem().getNode();
    }

    @Override
    public List<Integer> getValidErrors() {
        //TODO list of valid erros
        return null;
    }

    @Override
    public List<LockUsed> getUsedLocks() {
        //TODO list of locks
        return null;
    }

}
