package com.sos.jobnetmodel.graph;

import com.sos.JSHelper.Exceptions.JobSchedulerException;
import com.sos.graphviz.ClusterSubgraph;
import com.sos.graphviz.Edge;
import com.sos.graphviz.Graph;
import com.sos.graphviz.Node;
import com.sos.graphviz.enums.*;
import com.sos.jobnetmodel.objects.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class to create a graphical representation of a jobnet.
 *
 * This class use Graphviz for creating the jobnet.
 * User: ss
 * Date: 16.01.13
 * Time: 08:42
 */
public class GraphBuilder extends Graph {

    private final static Logger logger = LoggerFactory.getLogger(GraphBuilder.class);

    private final ImmutableJobnetNode rootChain;
    private final GraphOptions options;
    private final JobnetConverter converter;

    public GraphBuilder(JobnetConverter converter, GraphOptions options) {
        if(converter == null)
            throw new JobSchedulerException("The converter object should not be null.");

        if(!converter.isExecuted()) converter.run();
        this.rootChain = converter.getBootstrapNode();
        if(rootChain == null)
            throw new JobSchedulerException("The converter does not provide a bootstrap node - result of method getBootstrapNode is null.");

        this.options = options;
        this.converter = converter;
    }

    public void execute() {
        initGraph();
        JobnetContainer rootContainer = converter.getContainerList().get(rootChain.getInternalId());
        logger.info("The root container is " + rootContainer.getName());
        traverseJobNet(rootContainer, this);

        logger.info("Model  contains " + converter.getEdgeList().size() + " edges.");
        for(JobnetEdge e : converter.getEdgeList().values()) {
            logger.info("Edge: " + e.getFromNode().getInternalId() + "->" + e.getToNode().getInternalId());
            createEdge(e.getFromNode(),e.getToNode());
        }
    }

    public void traverseJobNet(JobnetContainer container, Graph parentGraph) {

        ClusterSubgraph nextGraph = createSubgraph(parentGraph, container.getName());
        for(ImmutableJobnetNode node : container.getNodes().values()) {

            if(node.getType() == JobnetNodeType.SUBNET) {
                logger.debug("Create SUBNET node " + node.getName());
                createChainNode(nextGraph, node);
                for(JobnetContainer c : container.getSuccessors().values()) {
                    traverseJobNet(c, nextGraph);
                }
            }

            if(node.getType() == JobnetNodeType.CONNECTOR) {
                logger.debug("Create CONNECTOR node " + node.getName());
                createConnectorNode(nextGraph,node);
                for(JobnetContainer c : container.getSuccessors().values()) {
                    traverseJobNet(c, nextGraph);
                }
            }

            if(node.getType() == JobnetNodeType.JOB) {
                logger.debug("Create JOB node " + node.getName());
                createJobNode(nextGraph, node);
            }

        }

    }

    private void initGraph() {
        init();
        getGlobalNodeProperties().setShape(Shape.Mrecord);
        getGlobalNodeProperties().setFixedSize(true);
        getGlobalNodeProperties().setWidth(7.5);
        getGlobalNodeProperties().setHeight(0.75);
        getGraphProperties().setDirection(options.getOrientation());
        getGraphProperties().setSplines(Splines.line);
        // getGraphProperties().setPackmode(PackMode.clust);
    }

    private ClusterSubgraph createSubgraph(Graph parentGraph, String name) {
        ClusterSubgraph nextGraph = parentGraph.newClusterSubgraph(name);
        nextGraph.getSubgraphProperties().setId(name);
        nextGraph.getSubgraphProperties().setLabel(name);
        // nextGraph.getSubgraphProperties().setStyle(Style.dotted);
        // nextGraph.getSubgraphProperties().setRank(RankType.source);
        nextGraph.getSubgraphProperties().setColor(SVGColor.gray);
        nextGraph.getSubgraphProperties().setLabeljust(Justify.l);
        nextGraph.getSubgraphProperties().setFontName("Arial");
        nextGraph.getSubgraphProperties().setFontSize(18);
        nextGraph.getSubgraphProperties().setFontColor(SVGColor.yellowgreen);
        nextGraph.getSubgraphProperties().setBgcolor(SVGColor.ivory);
        return nextGraph;
    }

    public Edge createEdge(JobnetNode from, JobnetNode to) {

        Node f = getNodeOrNullInAllGraphs(from.getInternalId());
        if (f == null) {
            // throw new RuntimeException("Node with id " + from.getInternalId() + " not present in Graph.");
            logger.warn("Node with id " + from.getInternalId() + " not present in Graph.");
            return null;
        }

        Node t = getNodeOrNullInAllGraphs(to.getInternalId());
        if (t == null) {
            // throw new RuntimeException("Node with id " + to.getInternalId() + " not present in Graph.");
            logger.warn("Node with id " + to.getInternalId() + " not present in Graph.");
            return null;
        }

        Edge e = newEdge(f, t);
        e.getEdgeProperties().setId(getEdgeId(e));
        if (options.getOrientation() == RankDir.TB) {
            e.getEdgeProperties().setHeadPort(PortPos.n);
            e.getEdgeProperties().setTailPort(PortPos.s);
        } else {
            e.getEdgeProperties().setHeadPort(PortPos.w);
            e.getEdgeProperties().setTailPort(PortPos.e);
        }
        return e;
    }

    public String getEdgeId(Edge e) {
        return e.getFromNode().getId() + "__" + e.getToNode().getId();
    }

    public Node createConnectorNode(Graph parentGraph, ImmutableJobnetNode jobnetNode) {
        logger.debug("Create CONNECTOR node with id {}", jobnetNode.getInternalId());
        Node node = parentGraph.newNode(jobnetNode.getInternalId());
        node.getSingleNodeProperties().setId(node.getId());
        //node.getSingleNodeProperties().setGroup(chain.getUniqueName());
        // String label = jobnetNodeImpl.getUniqueName();
        String label = jobnetNode.getUniqueName();
        node.getSingleNodeProperties().setLabel(label);
        node.getSingleNodeProperties().setShape(Shape.circle);
        node.getSingleNodeProperties().setHeight(0.3);
        node.getSingleNodeProperties().setWidth(0.3);
        setNodeAttributes(node, jobnetNode);
        return node;
    }

    public Node createJobNode(Graph parentGraph, ImmutableJobnetNode jobnetNode) {
        logger.debug("Create job node with id {}", jobnetNode.getInternalId());
        Node node = parentGraph.newNode(jobnetNode.getInternalId());
        node.getSingleNodeProperties().setId(node.getId());
        //node.getSingleNodeProperties().setGroup(chain.getUniqueName());
        // node.getSingleNodeProperties().setShape(Shape.box);

        String label = jobnetNode.getUniqueName();
        if(!jobnetNode.getProcessClass().isEmpty()) label += "| process_class=" + LabelBuilder.decode(jobnetNode.getProcessClass());
        node.getSingleNodeProperties().setLabel("{" + label + "}");

        /// node.getSingleNodeProperties().setLabel(node.getSingleNodeProperties().getGroup());
        setNodeAttributes(node, jobnetNode);
        return node;
    }

    public Node createChainNode(Graph parentGraph, ImmutableJobnetNode jobnetNode) {
        Node node = parentGraph.newNode(jobnetNode.getInternalId());
        node.getSingleNodeProperties().setId(node.getId());
        logger.debug("Create chain node with id {}", jobnetNode.getInternalId());

        String label = jobnetNode.getUniqueName();
        // label += "| " + LabelBuilder.decode(jobnetNode.getTitle());
        if(!jobnetNode.getProcessClass().isEmpty()) label += "| process_class=" + LabelBuilder.decode(jobnetNode.getProcessClass());
        node.getSingleNodeProperties().setLabel("{" + label + "}");

        // node.getSingleNodeProperties().setGroup(chain.getUniqueName());
        // node.getSingleNodeProperties().setShape(Shape.box);
        setNodeAttributes(node, jobnetNode);
        return node;
    }

    private void setNodeAttributes(Node node, ImmutableJobnetNode jobnetNode) {
        node.getSingleNodeProperties().setStyle("filled");
        node.getSingleNodeProperties().setColor(jobnetNode.getStatus().color(JobnetNodeType.SUBNET,PropertyName.color));
        node.getSingleNodeProperties().setFillcolor(jobnetNode.getStatus().color(JobnetNodeType.SUBNET,PropertyName.fillcolor));
        node.getSingleNodeProperties().setFontcolor(jobnetNode.getStatus().color(JobnetNodeType.SUBNET,PropertyName.fontcolor));
    }

}
