package com.sos.jobnetmodel.graph;

import com.sos.JSHelper.Exceptions.JobSchedulerException;
import com.sos.graphviz.enums.SVGColor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

/**
 * The properties for the jobnet nodes are loaded from a properties file given via the jvm property -Djobnet.graph.propertyFile.
 * See maven-surefire-plugin configuration in pom.xml for details.
 *
 * If this options is missing an the internal property file com/sos/jobnetmodel/properties/jobnode.properties with default settings is used.
 */
public class GraphNodeProperties {

    private final Logger logger = LoggerFactory.getLogger(GraphNodeProperties.class);

    private final static String nodePropertiesDefault = "/com/sos/jobnetmodel/graph/jobnode.properties";
    private String nodePropertiesResource = nodePropertiesDefault;

    private final static String graphPropertyFileOption = "jobnet.graph.propertyFile";

    private final String propertyPrefix;
    private Properties properties;

    public GraphNodeProperties(String propertyPrefix) {
        this.propertyPrefix = propertyPrefix;
        this.properties = loadProperties();
    }

    public void setProperties(String resourceName) {
        this.nodePropertiesResource = resourceName;
        this.properties = loadProperties();
    }

    private Properties loadProperties() {
        Properties result = new Properties();
        String propertiesFilename = System.getProperty(graphPropertyFileOption);
        if (propertiesFilename != null && !propertiesFilename.isEmpty()) {
            logger.info("Loading jobnode properties from file " + propertiesFilename);
            try {
                FileReader propertyFile = new FileReader(propertiesFilename);
                result.load(propertyFile);
            } catch (FileNotFoundException e) {
                throw new JobSchedulerException("File  " + propertiesFilename + " (given with jvm option jobnet.graph.propertyFile) does not exist.");
            } catch (IOException e) {
                throw new JobSchedulerException("Error loading the properties from file " + propertiesFilename);
            }
        } else {
            logger.info("Loading jobnode properties from resource " + nodePropertiesResource);
            try {
                InputStream in = GraphNodeProperties.class.getResourceAsStream(nodePropertiesResource);
                if (in == null)
                    throw new IOException();
                result.load(in);
            } catch (IOException e) {
                throw new JobSchedulerException("Error loading the properties from resource " + nodePropertiesResource);
            }
        }
        return result;
    }

    private String getPropertyName(PropertyName what) {
        return propertyPrefix + "." + what.name();
    }

    public String getValue(PropertyName what) {
        String name = getPropertyName(what);
        logger.debug("Reading property " + name);
        String result = properties.getProperty(name);
        return (result == null) ? "" : result;
    }

    public SVGColor getColorValue(PropertyName what, SVGColor defaultValue) {
        logger.debug("Get color value for property {}, default is {}.", what, defaultValue);
        SVGColor color = defaultValue;
        String colorString = getValue(what);
        if (!colorString.isEmpty()) {
            if(!SVGColor.isColor(colorString))
                throw new RuntimeException("The color value " + colorString + " is not valid.");
            color = SVGColor.valueOf(colorString);
        }
        logger.debug("The color value is " + color.name());
        return  color;
    }
}