package com.sos.jobnetmodel.objects;

import com.sos.JSHelper.Exceptions.JobSchedulerException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Class to convert an external jobnet representation in an internal object representation
 *
 * @Author ss
 * @version 23.01.13 11:42
 */
public class JobnetConverterImpl implements JobnetConverter {

    private final static Logger logger = LoggerFactory.getLogger(JobnetConverterImpl.class);

    private final ExternalJobnetNodeObject bootstrapNodeObject;
    private final String bootstrapNodeName;
    private final JobnetAdapter adapter;

    protected JobnetNode bootstrapNode;
    private JobnetContainer bootstrapContainer;
    protected final Map<String,JobnetNode> nodeList = new HashMap<String,JobnetNode>();
    private final Map<String,JobnetContainer> containerList = new HashMap<String,JobnetContainer>();
    private final Map<String,JobnetEdge> edgeList = new HashMap<String,JobnetEdge>();

    private List<String> edges = new ArrayList<String>();
    private List<String> pathes = new ArrayList<String>();

    protected boolean isExecuted = false;

    private int numberOfHeads = 0;
    private int numberOfConnectors = 0;
    private int numberOfNormalNodes = 0;

    public JobnetConverterImpl(JobnetAdapter adapter) {
        this.adapter = adapter;
        this.bootstrapNodeObject = adapter.getBootstrapNode();
        this.bootstrapNodeName = bootstrapNodeObject.getName();
    }

    public JobnetConverterImpl(ExternalJobnetNodeObject bootstrapNodeObject, String bootstrapNodeName) {
        this.adapter = null;
        this.bootstrapNodeObject = bootstrapNodeObject;
        this.bootstrapNodeName = bootstrapNodeName;
    }

    public void run() {
        bootstrapContainer = createContainer(null, bootstrapNodeName, bootstrapNodeObject.getInternalId());
        this.bootstrapNode = createChainNode(bootstrapNodeObject, bootstrapContainer);
        traverseJobNet(bootstrapNode, bootstrapContainer);

        for(JobnetNode n : nodeList.values()) {
            n.setConnector(nodeList.get(n.getExternalNodeObject().getConnector()));
            n.setSubnet(nodeList.get(n.getExternalNodeObject().getSubnet()));
        }

        // set child nodes for the representation as tree
        for(JobnetNode n : nodeList.values()) {
            n.setChildNodes(searchNodes(n.getType(),n.getUniqueName()));
        }

        isExecuted = true;
    }

    private Map<String,ImmutableJobnetNode> searchNodes(JobnetNodeType forType, String forName) {
        Map<String,ImmutableJobnetNode> result = new HashMap<String, ImmutableJobnetNode>();
        for(JobnetNode n : nodeList.values()) {
            if(n.getSubnet() != null && forType == JobnetNodeType.SUBNET && n.getType() == JobnetNodeType.CONNECTOR && n.getSubnet().getUniqueName().equalsIgnoreCase(forName))
                result.put(n.getUniqueName(),n);
            if(n.getConnector() != null && forType == JobnetNodeType.CONNECTOR && (n.getType() == JobnetNodeType.JOB || n.getType() == JobnetNodeType.SUBNET) && n.getConnector().getUniqueName().equalsIgnoreCase(forName))
                result.put(n.getUniqueName(),n);
        }
        return result;
    }

    public void traverseJobNet(JobnetNode jobnetNode, JobnetContainer parentContainer) {

        ExternalJobnetNodeObject nodeObject = jobnetNode.getExternalNodeObject();

        if (adapter == null)
            throw new JobSchedulerException("Use of class " + JobnetConverterImpl.class.getName() + " with constructor<ExternalJobnetNodeObject,String> requires to override the traverseJobNet method.");

        JobnetNode previousNode = null;
        JobnetNode currentNode = null;
        try {
            previousNode = getJobnetNode(nodeObject.getInternalId());
        } catch(Exception e) {
            logger.error("Error retrieving jobnet node for order " + nodeObject.getName());
            throw new JobSchedulerException(e);
        }

        logger.debug("process " + nodeObject.getSuccessors().size() + " successorNodes for node " + nodeObject.getInternalId());
        for (String key : nodeObject.getSuccessors()) {

            ExternalJobnetNodeObject node =  adapter.getNode(key);
            if(node != null) {
                if(nodeList.containsKey(node.getInternalId())) {
                    currentNode = nodeList.get(node.getInternalId());
                } else {
                    logger.debug("get node object for " + node.getInternalId());
                    currentNode = null;
                    logger.debug("type " + node.getType().name() + " detected.");

                    JobnetContainer nextContainer = parentContainer;
                    if (node.getType() == JobnetNodeType.CONNECTOR) {
                        if(!node.getUniqueName().startsWith(nextContainer.getName())) {
                            JobnetContainer c = parentContainer.getParent();
                            if (c!=null) nextContainer = c;
                        }
                        currentNode = createConnectorNode(node, nextContainer);
                    }

                    if (node.getType() == JobnetNodeType.JOB) {
                        currentNode = createJobNode(node, parentContainer);
                    }

                    if (node.getType() == JobnetNodeType.SUBNET) {
                        nextContainer = createContainer(parentContainer, node.getName(), node.getInternalId());
                        currentNode = createChainNode(node, nextContainer);
                    }
                    traverseJobNet(currentNode, nextContainer);
                }

                logger.debug("create edge from " + previousNode.getInternalId() + " to " + currentNode.getInternalId());
                createEdge(previousNode,currentNode);
            }

        }
    }

    public boolean isExecuted() {
        return isExecuted;
    }

    public JobnetContainer createContainer(JobnetContainer parent, String name, String id) {
        JobnetContainer c = new JobnetContainer(parent, name,id);
        logger.debug("Create container: {}/{}", id, name);
        containerList.put(id, c);
        if(parent != null)
            parent.addSuccessor(c);
        return c;
    }

    public JobnetEdge createEdge(JobnetNode from, JobnetNode to) {
        JobnetEdge edge = new JobnetEdge(from,to);
        String id = from.getInternalId() + "__" + to.getInternalId();
        edgeList.put(id,edge);
        // logger.debug("Create edge with id {}",id);
        from.getSuccessorNodes().put(to.getInternalId(),to);
        to.getPredeceessorNodes().put(from.getInternalId(),from);
        return edge;
    }

    /**
     * Creates a node for a CONNECTOR. A CONNECTOR represents always a step of a cronacle chain, but it is necessary only
     * the step itself and its previous step has more than one element. The name of the node is build by concat the name
     * of the parent chain and the (internal) id of the step.
     *
     * @param stepObject
     * @param container
     * @return
     */
    public JobnetNode createConnectorNode(ExternalJobnetNodeObject stepObject, JobnetContainer container) {
        // logger.debug("create CONNECTOR: " + getUniqueName(stepObject) + "/" + stepObject.getInternalId());
        numberOfConnectors++;
        return createJobNetNode(container, JobnetNodeType.CONNECTOR, stepObject);
    }

    /**
     * Creates a node for a single job. Because the script name is not unique the node id will be appended.
     *
     * @param jobObject
     * @param container
     * @return
     */
    public JobnetNode createJobNode(ExternalJobnetNodeObject jobObject, JobnetContainer container) {
        numberOfNormalNodes++;
        return createJobNetNode(container, JobnetNodeType.JOB, jobObject);
    }

    /**
     * Creates a node for a chain assuming the name of the chain is unique.
     *
     * @param chainObject
     * @param container
     * @return
     */
    public JobnetNode createChainNode(ExternalJobnetNodeObject chainObject, JobnetContainer container) {
        numberOfHeads++;
        JobnetNode result = createJobNetNode(container, JobnetNodeType.SUBNET, chainObject);
        return createJobNetNode(container, JobnetNodeType.SUBNET, chainObject);
    }

    /**
     * A generic method to create a JobnetNode (Chain, Job or Connector).
     *
     * @param container
     * @param type
     * @param nodeObject
     * @return
     */
    public JobnetNode createJobNetNode(JobnetContainer container, JobnetNodeType type, ExternalJobnetNodeObject nodeObject) {
        // logger.info("Create node: " + uniqueName + " in container " + container.getInternalId() + "/" + externalNodeObject.getInternalId());
        String uniqueName = getUniqueName(nodeObject);
        JobnetNode node = nodeObject.getJobnetNodeObject();
        if (node == null) {
            node = new JobnetNodeImpl(nodeObject.getInternalId(), type);
        }
        node.setUniqueName(uniqueName);
        node.setName(getName(nodeObject));
        node.setNodeStatus(node.getStatus());
        node.setTitle(nodeObject.getTitle());
        node.setValidErrors(nodeObject.getValidErrors());
        logger.info("Put node " + node.getName() + " with id " + node.getInternalId() + " to nodelist.");
        node.setContainer(container);
        nodeList.put(node.getInternalId(), node);
        node.setExternalNodeObject(nodeObject);
        if (container != null) {
            JobnetContainer c = containerList.get(container.getId());
            c.addNode(node);
        }
        return node;
    }

    @Override
    public String getUniqueName(ExternalJobnetNodeObject node) {
        return node.getUniqueName();
    }

    @Override
    public String getName(ExternalJobnetNodeObject node) {
        return node.getName();
    }

    @Override
    public Map<String,JobnetContainer> getContainerList() {
        return containerList;
    }

    @Override
    public Map<String,JobnetNode> getNodeList() {
        return nodeList;
    }

    @Override
    public Map<String,JobnetEdge> getEdgeList() {
        return edgeList;
    }

    @Override
    public JobnetContainer getBootstrapContainer() {
        return bootstrapContainer;
    }

    @Override
    public String getBootstrapNodeName() {
        return bootstrapNodeName;
    }

    @Override
    public boolean isUniqueName(ExternalJobnetNodeObject externalNode, String baseName, int index) {
        boolean result = true;
        String searchFor = getIndexedName(baseName,index);
        for(JobnetNode node : nodeList.values()) {
            if(node.getUniqueName().equals(searchFor) && node.getInternalId() != externalNode.getInternalId()) {
                result = false;
                break;
            }
        }
        return result;
    }

    @Override
    public String getIndexedName(String baseName, int index) {
        String result = (index > 0) ? baseName + "_"  + String.valueOf(index) : baseName;
        return result.toLowerCase();
    }

    public ExternalJobnetNodeObject getBootstrapNodeObject() {
        return bootstrapNodeObject;
    }

    public JobnetNode getBootstrapNode() {
        return bootstrapNode;
    }

    public JobnetNode getJobnetNode(String key) {
        if(!getNodeList().containsKey(key))
            throw new JobSchedulerException("A node with ID " + key + " is not present in the nodelist.");
        return getNodeList().get(key);
    }

    @Override
    public List<String> getEdges() {
        if (!isExecuted())
            logger.warn("The list of the jobnet edges is not available before the jobnet is built with the run() method.");
        edges.clear();
        for (JobnetEdge edge : getEdgeList().values() )
            edges.add(edge.getFromNode().getInternalId() + " -> " + edge.getToNode().getInternalId());
        return edges;
    }

    @Override
    public List<String> getPathes() {
        if (!isExecuted())
            logger.warn("The list of the jobnet pathes is not available before the jobnet is built with the run() method.");
        pathes.clear();
        buildPathes(getBootstrapNode(), "");
        return pathes;
    }

    @Override
    public int numberOfContainer() {
        return getContainerList().size();
    }

    @Override
    public int numberOfHeads() {
        return numberOfHeads;
    }

    @Override
    public int numberOfConnectores() {
        return numberOfConnectors;
    }

    @Override
    public int numberOfNormalNodes() {
        return numberOfNormalNodes;
    }

    private void buildPathes(ImmutableJobnetNode node, String path) {

        StringBuffer result = new StringBuffer();
        result.append(path);
        if(result.length() > 0)
            result.append(" -> ");
        result.append(node.getInternalId());

        if(!node.hasSuccessorNodes()) {
            pathes.add(result.toString());
            logger.debug(result.toString());
        } else {
            for(String id : node.getSuccessorNodes().keySet()) {
                ImmutableJobnetNode order = getNodeList().get(id);
                buildPathes(order, result.toString());
            }
        }

    }

}
