package com.sos.jobnetmodel.objects;

import com.sos.JSHelper.Exceptions.JobSchedulerException;
import com.sos.graphviz.enums.SVGColor;
import com.sos.jobnetmodel.graph.GraphNodeProperties;
import com.sos.jobnetmodel.graph.PropertyName;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.Map;

/**
 * An enumeration for all possible states of a jobnet node.
 */
public enum NodeStatus {
    NOT_PROCESSED       (0,  40, SVGColor.gray),
    RUNNING             (1,  50, SVGColor.gold),
    FINISHED            (2,  30, SVGColor.yellowgreen),
    ERROR               (3, 100, SVGColor.red),
    IGNORED             (4,  20, SVGColor.whitesmoke),
    WAITING             (5,  80, SVGColor.steelblue),
    DISPATCHING         (6,  60, SVGColor.sienna),
    STARTED             (8,  70, SVGColor.goldenrod),
    RESTART_PREPARED    (9,  10, SVGColor.silver),
    KILLED              (10, 90, SVGColor.crimson),
    ABORTED             (11, 89, SVGColor.tomato),
    UNDEFINED           (99,  5, SVGColor.magenta);

    // Rank: ERROR, KILLED, ABORTED, WAITING, STARTED, DISPATCHING, RUNNING, NOT_PROCESSED, FINISHED, IGNORED, RESTART_PREPARED, UNDEFINED

    private final Logger logger = LoggerFactory.getLogger(NodeStatus.class);

    private final int status;
    private int statusRank;
    private final SVGColor defaultColor;
    private NodeStatusText text;

    private Map<JobnetNodeType,GraphNodeProperties> nodePropertiesList = new HashMap<JobnetNodeType,GraphNodeProperties>();

    private String nodePropertyResource = null;
    private static Map<Integer, NodeStatus> map = new HashMap<Integer, NodeStatus>();
    static {
        for (NodeStatus state : NodeStatus.values()) {
            map.put(state.status, state);
        }
    }

    private NodeStatus(int status, int statusRank, SVGColor defaultColor) {
        this.status = status;
        this.statusRank = statusRank;
        this.defaultColor = defaultColor;
        this.text = new NodeStatusText(this);
    }

    public Integer getIndex() {
        return status;
    }

    public SVGColor color(JobnetNodeType forType, PropertyName propertyName) {
        logger.info("Try to get color information for type {}, property {}", forType, propertyName);
        String identifier = normalize(name() + "." + normalize(forType.name()) );   // e.g. error.subnet or finished.connector
        logger.trace("The identifier is {}.",identifier);

        GraphNodeProperties nodeProperties = nodePropertiesList.get(forType);
        if(nodeProperties == null) {
            logger.info("Load the node properties.");
            nodeProperties = new GraphNodeProperties(identifier);
            if (nodePropertyResource != null)
                nodeProperties.setProperties(nodePropertyResource);
            nodePropertiesList.put(forType,nodeProperties);
        }
        return nodeProperties.getColorValue(propertyName, defaultColor);
    }

    private String normalize(String name) {
        return name.toLowerCase();
    }

    public void setNodeProperties(String resourceName) {
        nodePropertyResource = resourceName;
    }

    public static NodeStatus valueOf(int valueOfState) {
        NodeStatus result = map.get(valueOfState);
        if (result == null)
            throw new JobSchedulerException("The index " + valueOfState + " does not point to a valid NodeStatus.");
        return map.get(valueOfState);
    }

    public boolean isRunning() {
        return(this == NodeStatus.RUNNING);
    }

    public boolean isNotProcessed() {
        return(this == NodeStatus.NOT_PROCESSED);
    }

    public boolean isRestartPrepared() {
        return(this == NodeStatus.RESTART_PREPARED);
    }

    public boolean isFinished() {
        return(this == NodeStatus.FINISHED);
    }

    public boolean isStarted() {
        return(this == NodeStatus.STARTED);
    }

    public boolean isError() {
        return(this == NodeStatus.ERROR);
    }

    public boolean isIgnored() {
        return(this == NodeStatus.IGNORED);
    }

    public boolean isWaiting() {
        return(this == NodeStatus.WAITING);
    }

    public boolean isDispatching() {
        return(this == NodeStatus.DISPATCHING);
    }

    public String getMsg() {
        return text.getMsg();
    }

    public int getStatusRank() {
        return statusRank;
    }

}

