package com.sos.scheduler.engine.kernel

import Scheduler._
import com.google.common.base.Objects.firstNonNull
import com.google.common.io.Closer
import com.google.inject.Guice.createInjector
import com.google.inject.Injector
import com.sos.scheduler.engine.common.async.CallRunner
import com.sos.scheduler.engine.common.inject.GuiceImplicits._
import com.sos.scheduler.engine.common.log.LoggingFunctions.enableJavaUtilLoggingOverSLF4J
import com.sos.scheduler.engine.common.scalautil.HasCloser.implicits._
import com.sos.scheduler.engine.common.scalautil.Logger
import com.sos.scheduler.engine.common.xml.NamedChildElements
import com.sos.scheduler.engine.common.xml.XmlUtils.{childElements, loadXml}
import com.sos.scheduler.engine.cplusplus.runtime.annotation.ForCpp
import com.sos.scheduler.engine.cplusplus.runtime.{CppProxy, CppProxyInvalidatedException, DisposableCppProxyRegister, Sister}
import com.sos.scheduler.engine.data.filebased.{FileBasedEvent, FileBasedType}
import com.sos.scheduler.engine.data.log.SchedulerLogLevel
import com.sos.scheduler.engine.data.scheduler.SchedulerCloseEvent
import com.sos.scheduler.engine.eventbus.{EventSubscription, SchedulerEventBus}
import com.sos.scheduler.engine.kernel.async.SchedulerThreadFutures.inSchedulerThread
import com.sos.scheduler.engine.kernel.async.{CppCall, SchedulerThreadCallQueue}
import com.sos.scheduler.engine.kernel.command.{CommandSubsystem, UnknownCommandException}
import com.sos.scheduler.engine.kernel.configuration.SchedulerModule
import com.sos.scheduler.engine.kernel.cppproxy.SpoolerC
import com.sos.scheduler.engine.kernel.database.DatabaseSubsystem
import com.sos.scheduler.engine.kernel.event.EventSubsystem
import com.sos.scheduler.engine.kernel.filebased.FileBasedSubsystem
import com.sos.scheduler.engine.kernel.log.{CppLogger, PrefixLog}
import com.sos.scheduler.engine.kernel.plugin.{PluginModule, PluginSubsystem}
import com.sos.scheduler.engine.kernel.scheduler._
import com.sos.scheduler.engine.kernel.security.SchedulerSecurityLevel
import com.sos.scheduler.engine.kernel.time.TimeZones
import com.sos.scheduler.engine.kernel.util.MavenProperties
import com.sos.scheduler.engine.main.SchedulerControllerBridge
import java.lang.Thread.currentThread
import javax.annotation.Nullable
import javax.inject.{Inject, Singleton}
import org.joda.time.DateTimeZone.UTC
import org.joda.time.Instant.now
import scala.collection.JavaConversions._
import scala.collection.breakOut
import scala.util.control.NonFatal

@ForCpp
@Singleton
final class Scheduler @Inject private(
    cppProxy: SpoolerC,
    controllerBridge: SchedulerControllerBridge,
    schedulerConfiguration: SchedulerConfiguration,
    prefixLog: PrefixLog,
    disposableCppProxyRegister: DisposableCppProxyRegister,
    implicit private val schedulerThreadCallQueue: SchedulerThreadCallQueue,
    eventBus: SchedulerEventBus,
    val injector: Injector)
extends Sister
with SchedulerIsClosed
with SchedulerXmlCommandExecutor
with HasInjector {

  private var closed = false
  private val closer = Closer.create()
  private val callRunner = new CallRunner(schedulerThreadCallQueue.delegate)
  private lazy val pluginSubsystem = injector.apply[PluginSubsystem]
  private lazy val commandSubsystem = injector.apply[CommandSubsystem]
  private lazy val databaseSubsystem = injector.apply[DatabaseSubsystem]

  val startInstant = now()

  enableJavaUtilLoggingOverSLF4J()
  TimeZones.initialize()
  //DateTimeZone.setDefault(UTC);
  java.util.TimeZone.setDefault(UTC.toTimeZone)       // Für JPA @Temporal(TIMESTAMP), damit Date wirklich UTC enthält. Siehe http://stackoverflow.com/questions/508019
  Thread.currentThread.setContextClassLoader(getClass.getClassLoader)   // Für Mail, http://stackoverflow.com/questions/1969667

  cppProxy.setSister(this)

  if (controllerBridge eq EmptySchedulerControllerBridge.singleton) { // Wenn wir ein controllerBridge haben, ist der Scheduler über Java (CppScheduler.main) aufgerufen worden. Dort wird die Sperre gesetzt.
    threadLock()
    closer { threadUnlock() } //TODO Sperre wird in onClose() zu früh freigegeben, der Scheduler läuft ja noch. Lösung: Start über Java mit CppScheduler.run()
  }

  @ForCpp
  private def initialize() {
    val eventSubscription = {
      val subsystemDescriptions = injector.apply[FileBasedSubsystem.Register].descriptions
      val subsystemMap: Map[FileBasedType, FileBasedSubsystem] = subsystemDescriptions.map { o ⇒ o.fileBasedType -> injector.getInstance(o.subsystemClass) }(breakOut)
      EventSubscription[FileBasedEvent] { e ⇒ for (subsystem <- subsystemMap.get(e.typedPath.fileBasedType)) subsystem.onFileBasedEvent(e) }
    }
    eventBus.registerHot(eventSubscription)
    closer { eventBus.unregisterHot(eventSubscription) }
  }

  def onCppProxyInvalidated() {}

  @ForCpp private def onClose() {
    closed = true
    try {
      schedulerThreadCallQueue.close()
      eventBus.publish(new SchedulerCloseEvent)
      eventBus.dispatchEvents()
      schedulerThreadCallQueue.close()
      try databaseSubsystem.close() catch { case NonFatal(x) ⇒ prefixLog.error(s"databaseSubsystem.close(): $x") }
      try pluginSubsystem.close() catch { case NonFatal(x) ⇒ prefixLog.error(s"pluginSubsystem.close(): $x") }
    }
    eventBus.dispatchEvents()
    disposableCppProxyRegister.tryDisposeAll()
    closer.close()
  }

  @ForCpp private def onLoad() {
    pluginSubsystem.initialize()
    controllerBridge.onSchedulerStarted(this)
  }

  @ForCpp private def onActivate() {
    pluginSubsystem.activate()
  }

  @ForCpp private def onActivated() {
    controllerBridge.onSchedulerActivated()
  }

  /** Wird bei jedem Schleifendurchlauf aufgerufen. */
  @ForCpp private def onEnteringSleepState(): Long = {
    val somethingDone = callRunner.executeMatureCalls()
    eventBus.dispatchEvents()
    val nextTime = schedulerThreadCallQueue.nextTime
    if (somethingDone) -nextTime else nextTime
  }

  /** Nur für C++, zur Ausführung eines Kommandos in Java */
  @ForCpp private def javaExecuteXml(xml: String): String = {
    try commandSubsystem.executeXml(xml)
    catch {
      case x: UnknownCommandException ⇒
        prefixLog.warn(x.toString)
        "UNKNOWN_COMMAND"   // Siehe command_error.cxx, für ordentliche Meldung SCHEDULER-105, bis Java die selbst liefert kann.
    }
  }

  @ForCpp private def instance[A](c: Class[A]): A =
    injector.getInstance(c)

  @ForCpp private def getEventSubsystem =
    injector.getInstance(classOf[EventSubsystem])

  @ForCpp private def log(prefix: String, level: Int, line: String) {
    CppLogger.log(prefix, SchedulerLogLevel.ofCpp(level), line)
  }

  @ForCpp private def enqueueCall(o: CppCall) {
    schedulerThreadCallQueue.add(o)
  }

  @ForCpp private def cancelCall(o: CppCall) {
    schedulerThreadCallQueue.tryCancel(o)
  }

  @ForCpp private def threadLock() {
    CppProxy.threadLock.lock()
  }

  @ForCpp /*private*/ def threadUnlock() {
    CppProxy.threadLock.unlock()
  }

  def terminate() {
    if (!isClosed) {
      try cppProxy.cmd_terminate()
      catch {
        case x: CppProxyInvalidatedException ⇒
          logger.debug("Scheduler.terminate() ignored because C++ object has already been destroyed")
      }
    }
  }

  /** Löst bei einem ERROR-Element eine Exception aus. */
  def executeXml(xml: String): String = {
    val result = uncheckedExecuteXml(xml)
    if (result contains "<ERROR") {
      for (e <- childElements(loadXml(result).getDocumentElement);
           error <- new NamedChildElements("ERROR", e))
        throw new SchedulerException(error.getAttribute("text"))
    }
    result
  }

  /** execute_xml_string() der C++-Klasse Spooler */
  def uncheckedExecuteXml(xml: String): String = {
    if (closed) sys.error("Scheduler is closed")
    inSchedulerThread { cppProxy.execute_xml_string(xml) }
    .stripSuffix("\0")  // Von C++ angehängtes '\0' an, siehe Command_response::end_standard_response()
  }

  def uncheckedExecuteXml(xml: String, securityLevel: SchedulerSecurityLevel, clientHostName: String) =
    inSchedulerThread { cppProxy.execute_xml_string_with_security_level(xml, securityLevel.cppName, clientHostName) }
    .stripSuffix("\0")  // Von C++ angehängtes '\0' an, siehe Command_response::end_standard_response()

  //    /** @param text Sollte auf \n enden */
  //    public void writeToSchedulerLog(LogCategory category, String text) {
  //        cppProxy.write_to_scheduler_log(category.string(), text);
  //    }

  def callCppAndDoNothing() {
    cppProxy.tcp_port
  }

  def overview: SchedulerOverview =
    inSchedulerThread {
      new SchedulerOverview(
        version = mavenProperties.version,
        versionCommitHash = mavenProperties.versionCommitHash,
        startInstant = startInstant,
        instant = now(),
        schedulerId = schedulerConfiguration.schedulerId,
        tcpPort = schedulerConfiguration.tcpPort match { case 0 ⇒ None case n ⇒ Some(n) },
        udpPort = schedulerConfiguration.udpPort,
        processId = cppProxy.pid,
        state = cppProxy.state_name)
    }

  def isClosed = closed
}

@ForCpp
object Scheduler {
  private val logger = Logger(getClass)
  private val mavenProperties = MavenProperties(getClass)

  @ForCpp def of(cppProxy: SpoolerC, @Nullable controllerBridgeOrNull: SchedulerControllerBridge, configurationXml: String) = {
    val controllerBridge = firstNonNull(controllerBridgeOrNull, EmptySchedulerControllerBridge.singleton)
    controllerBridge.cppSettings.setSettingsInCpp(cppProxy.modifiable_settings)

    val injector = createInjector(Seq(
      new SchedulerModule(cppProxy, controllerBridge, schedulerThread = currentThread),
      PluginModule(configurationXml)))
    injector.getInstance(classOf[Scheduler])
  }

  @ForCpp
  def buildVersion: String =
    try mavenProperties.buildVersion
    catch { case e: NoSuchElementException ⇒ "" }

  @ForCpp
  def versionCommitHash: String =
    if (mavenProperties.version endsWith "-SNAPSHOT")
      try mavenProperties.versionCommitHash
      catch { case e: NoSuchElementException ⇒ "" }
    else
      ""
}
